

/*
   G. Wiesenekker. ZZZZZZ a chess program.
   Copyright (C) 1993-1994 G. Wiesenekker
   E-mail: wiesenekker@sara.nl

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "zzzzzz.h"

#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL 0
#endif

#ifdef _MSC_VER
#define inline __inline
#define CDECL __cdecl
#else
#define CDECL
#endif



char           *
                time_stamp(void)
{
    time_t          t;

    t = time(NULL);
    return (ctime(&t));
}

/*
   my_printf works just like printf, but optionally echoes to a file
   (and flushes the buffer, so that you can still see what went wrong
   just before the core dump)
 */

void            my_printf(char *format,...)
{
    va_list         arg;


    va_start(arg, format);
    vfprintf(stdout, format, arg);
    fflush(stdout);


    va_end(arg);
}

/*
   my_fgets does an fgets, but removes the newline.
 */

char           *
                my_fgets(char *s, int n, FILE * stream)
{
    char           *p;

    if (fgets(s, n, stream) == NULL)
        return (NULL);
    for (p = s; *p != '\0'; p++)
        if (*p == '\n')
            *p = '\0';
    return (s);
}

/*
   A 32 bit random number generator. An implementation in C of the algorithm given by
   Knuth, the art of computer programming, vol. 2, pp. 26-27. We use e=32, so
   we have to evaluate y(n) = y(n - 24) + y(n - 55) mod 2^32, which is implicitly
   done by unsigned arithmetic.
 */

ulong_t
ulrand(void)
{
    /* random numbers from Mathematica 2.0. SeedRandom = 1;
     * Table[Random[Integer, {0, 2^32 - 1}] */
    static ulong_t  x[55] =
    {
        1410651636UL,
        3012776752UL,
        3497475623UL,
        2892145026UL,
        1571949714UL,
        3253082284UL,
        3489895018UL,
        387949491UL,
        2597396737UL,
        1981903553UL,
        3160251843UL,
        129444464UL,
        1851443344UL,
        4156445905UL,
        224604922UL,
        1455067070UL,
        3953493484UL,
        1460937157UL,
        2528362617UL,
        317430674UL,
        3229354360UL,
        117491133UL,
        832845075UL,
        1961600170UL,
        1321557429UL,
        747750121UL,
        545747446UL,
        810476036UL,
        503334515UL,
        4088144633UL,
        2824216555UL,
        3738252341UL,
        3493754131UL,
        3672533954UL,
        29494241UL,
        1180928407UL,
        4213624418UL,
        33062851UL,
        3221315737UL,
        1145213552UL,
        2957984897UL,
        4078668503UL,
        2262661702UL,
        65478801UL,
        2527208841UL,
        1960622036UL,
        315685891UL,
        1196037864UL,
        804614524UL,
        1421733266UL,
        2017105031UL,
        3882325900UL,
        810735053UL,
        384606609UL,
        2393861397UL
    };
    static int      init = true;
    static ulong_t  y[55];
    static int      j,
                    k;
    ulong_t         ul;

    if (init) {
        int             i;

        init = false;
        for (i = 0; i < 55; i++)
            y[i] = x[i];
        j = 24 - 1;
        k = 55 - 1;
    }
    ul = (y[k] += y[j]);
    if (--j < 0)
        j = 55 - 1;
    if (--k < 0)
        k = 55 - 1;
    return (ul);
}


ulong_t
crc(uchar_t * u)
{
    ulong_t         crc;
    static ulong_t  crctab[256] =
    {
        0UL,
        79764919UL,
        159529838UL,
        222504665UL,
        319059676UL,
        398814059UL,
        445009330UL,
        507990021UL,
        638119352UL,
        583659535UL,
        797628118UL,
        726387553UL,
        890018660UL,
        835552979UL,
        1015980042UL,
        944750013UL,
        1276238704UL,
        1221641927UL,
        1167319070UL,
        1095957929UL,
        1595256236UL,
        1540665371UL,
        1452775106UL,
        1381403509UL,
        1780037320UL,
        1859660671UL,
        1671105958UL,
        1733955601UL,
        2031960084UL,
        2111593891UL,
        1889500026UL,
        1952343757UL,
        2552477408UL,
        2632100695UL,
        2443283854UL,
        2506133561UL,
        2334638140UL,
        2414271883UL,
        2191915858UL,
        2254759653UL,
        3190512472UL,
        3135915759UL,
        3081330742UL,
        3009969537UL,
        2905550212UL,
        2850959411UL,
        2762807018UL,
        2691435357UL,
        3560074640UL,
        3505614887UL,
        3719321342UL,
        3648080713UL,
        3342211916UL,
        3287746299UL,
        3467911202UL,
        3396681109UL,
        4063920168UL,
        4143685023UL,
        4223187782UL,
        4286162673UL,
        3779000052UL,
        3858754371UL,
        3904687514UL,
        3967668269UL,
        881225847UL,
        809987520UL,
        1023691545UL,
        969234094UL,
        662832811UL,
        591600412UL,
        771767749UL,
        717299826UL,
        311336399UL,
        374308984UL,
        453813921UL,
        533576470UL,
        25881363UL,
        88864420UL,
        134795389UL,
        214552010UL,
        2023205639UL,
        2086057648UL,
        1897238633UL,
        1976864222UL,
        1804852699UL,
        1867694188UL,
        1645340341UL,
        1724971778UL,
        1587496639UL,
        1516133128UL,
        1461550545UL,
        1406951526UL,
        1302016099UL,
        1230646740UL,
        1142491917UL,
        1087903418UL,
        2896545431UL,
        2825181984UL,
        2770861561UL,
        2716262478UL,
        3215044683UL,
        3143675388UL,
        3055782693UL,
        3001194130UL,
        2326604591UL,
        2389456536UL,
        2200899649UL,
        2280525302UL,
        2578013683UL,
        2640855108UL,
        2418763421UL,
        2498394922UL,
        3769900519UL,
        3832873040UL,
        3912640137UL,
        3992402750UL,
        4088425275UL,
        4151408268UL,
        4197601365UL,
        4277358050UL,
        3334271071UL,
        3263032808UL,
        3476998961UL,
        3422541446UL,
        3585640067UL,
        3514407732UL,
        3694837229UL,
        3640369242UL,
        1762451694UL,
        1842216281UL,
        1619975040UL,
        1682949687UL,
        2047383090UL,
        2127137669UL,
        1938468188UL,
        2001449195UL,
        1325665622UL,
        1271206113UL,
        1183200824UL,
        1111960463UL,
        1543535498UL,
        1489069629UL,
        1434599652UL,
        1363369299UL,
        622672798UL,
        568075817UL,
        748617968UL,
        677256519UL,
        907627842UL,
        853037301UL,
        1067152940UL,
        995781531UL,
        51762726UL,
        131386257UL,
        177728840UL,
        240578815UL,
        269590778UL,
        349224269UL,
        429104020UL,
        491947555UL,
        4046411278UL,
        4126034873UL,
        4172115296UL,
        4234965207UL,
        3794477266UL,
        3874110821UL,
        3953728444UL,
        4016571915UL,
        3609705398UL,
        3555108353UL,
        3735388376UL,
        3664026991UL,
        3290680682UL,
        3236090077UL,
        3449943556UL,
        3378572211UL,
        3174993278UL,
        3120533705UL,
        3032266256UL,
        2961025959UL,
        2923101090UL,
        2868635157UL,
        2813903052UL,
        2742672763UL,
        2604032198UL,
        2683796849UL,
        2461293480UL,
        2524268063UL,
        2284983834UL,
        2364738477UL,
        2175806836UL,
        2238787779UL,
        1569362073UL,
        1498123566UL,
        1409854455UL,
        1355396672UL,
        1317987909UL,
        1246755826UL,
        1192025387UL,
        1137557660UL,
        2072149281UL,
        2135122070UL,
        1912620623UL,
        1992383480UL,
        1753615357UL,
        1816598090UL,
        1627664531UL,
        1707420964UL,
        295390185UL,
        358241886UL,
        404320391UL,
        483945776UL,
        43990325UL,
        106832002UL,
        186451547UL,
        266083308UL,
        932423249UL,
        861060070UL,
        1041341759UL,
        986742920UL,
        613929101UL,
        542559546UL,
        756411363UL,
        701822548UL,
        3316196985UL,
        3244833742UL,
        3425377559UL,
        3370778784UL,
        3601682597UL,
        3530312978UL,
        3744426955UL,
        3689838204UL,
        3819031489UL,
        3881883254UL,
        3928223919UL,
        4007849240UL,
        4037393693UL,
        4100235434UL,
        4180117107UL,
        4259748804UL,
        2310601993UL,
        2373574846UL,
        2151335527UL,
        2231098320UL,
        2596047829UL,
        2659030626UL,
        2470359227UL,
        2550115596UL,
        2947551409UL,
        2876312838UL,
        2788305887UL,
        2733848168UL,
        3165939309UL,
        3094707162UL,
        3040238851UL,
        2985771188UL
    };


    for (crc = 0; *u != '\0'; u++)
        crc = (crc << 8) ^ crctab[(crc >> 24) ^ *u];

    return (crc);
}


/*
   read_game reads the moves played sofar
   and returns the number of moves done by white and by black.
 
*/


static int       read_game(FILE * f)
{
    char            line[LINE_MAX];
    char            move[LINE_MAX];
    int             iroot;

    nwhite = nblack = 0;
   
   
   
    while (my_fgets(line, LINE_MAX, f) != NULL) {
       
          
        if (sscanf(line, "%s", move) != 1)
            continue;
        
        gen_root();
        BUG((iroot = search_root(move, MOVE_STRING_ANY, false, true)) == INVALID)
            do_move(root[iroot].root_move);
       
        if (your_colour == WHITE_COLOUR)
            nwhite++;
        else
            nblack++;
    }
    compact_lists();

     if (my_colour == WHITE_COLOUR)
        return (nwhite);
    else
        return (nblack);  
     
    
}


/*
   play_game plays a game under tournament conditions.
   zzzzzz plays xxxxxx, you play yyyyyy.
 */

static void      play_game(int xxxxxx, int yyyyyy, char *name)
{
   
    FILE           *f;
    time_t          wall1,
                    wall2;
    int             wall;
    int             iroot;
    int n;
    char            string[LINE_MAX];
    char            line[LINE_MAX];
    char            move[LINE_MAX];
    char buf[256*4], buf2[30], buf3[30], buf4[30], buf5[30],
         buf6[30], buf7[30], buf8[30], buf9[30], buf10[30],
         buf11[30], buf12[30];
         
         my_colour = NO_COLOUR;
    
    
   DUMMY("%d", yyyyyy);  
    
    for (;;) {                          
    
    
    fen_to_board(fen, true);     
    BUG((f = fopen(name, "r")) == NULL);
    ngame_moves = read_game(f);
    FCLOSE(f);
    
     if (using_setboard) {    
     compact_lists();
}
       
      
        if (draw_by_repetition(true)) {
             my_printf("draw by rep!\n");
             printf("1/2-1/2 {draw by repetition}\n");                        
        }
      
          
   gen_root();
              
        if (nroot == 0) {
            
                 if (i_am_check(my_colour, your_colour) && !analyze_mode) {                 
                 my_printf("%sis mate!\n", colour_as_string[my_colour]);
                 printf("computer mates\n");
                 goto start;
                 }else
                 if (analyze_mode) {
                 if (i_am_check(my_colour, your_colour) || i_am_check(your_colour, my_colour)) { 
                 
                 printf(" >>> Shortest Mate Found <<< ");
                 }      
              
            }else
            
                 if (!analyze_mode) {
                 my_printf("stale mate!\n");
                 printf("1/2-1/2 {stalemate}\n");
                 goto start;         
                 }   
              
     
     
     
     
        } else {
                      
            
            if (xxxxxx == my_colour && !force_mode) {
               
                if (nroot == 1) {
                    my_printf("\nforced!\n");
                    jroot = 0;
                    root[jroot].root_score = 0;
                   
                } else {
              
                
                          
                    wall1 = time(NULL);
                      
                    // look for shortest mate first - JA
                    
                    if (nroot == 0){
                       DepthMax = 32;  
                       do_mate();                
                       }
                       DepthMax = 32;      
                       do_root(); 
                                                                               
          
                   
                    
                    wall2 = time(NULL);
                    wall = (int) difftime(wall2, wall1) + 1;
                    GameTimeUsed += wall;
                    my_printf("\nWallTime=%d GameTimeUsed=%d\n", wall, GameTimeUsed);
                }
               
        
              
                                         
                if  (Best_Score > -RESIGN_THRESHOLD) {
                resigncounter = 0;
                }else{                                      
                if (resigncounter > 1) printf("zzzzzz thinking about resigning ..\n");
                resigncounter++;                
                }
                if (resigncounter == 5) {
                printf("\nzzzzzz resigns\n\n");
                printf("\ncomputer resigns\n");
                force_mode = true;
              //  continue;
			} 
           
                                  
                move_to_san_string(root[jroot].root_move, true);      
                printf("\n");
                printf("\move %s", move_string);
                printf("\n");
                             
          
                BUG((f = fopen(name, "a")) == NULL);
                fprintf(f, "%7s zzzzzz %.3f\n", move_string, (double) root[jroot].root_score);
                FCLOSE(f);
                          
               
             continue;       
            }
            
            start:
                           
                                                                                                                                                                                                       
               if (!fgets(line, 256, stdin)) {
                return;
	           	}

	          	if (line[0] == '\n') {
		       	continue;
	          	}
                                                                  
                       sscanf(line, "%s", move);        
                    
                       if (!strcmp(move, "go")) {
                       force_mode = false;
                       analyze_mode = false;
                       xxxxxx = my_colour;                                        
                       play_game(my_colour, your_colour, "game.txt");     
                       continue;
                       }
                                                                         
                       if (!strcmp(move, "quit") || !strcmp(move, "exit")) {
                       if (analyze_on) { 
                       analyze_on = false; 
                       analyze_mode = false;
                       opponent = true;
                       continue;
                       }else{                    
                       printf("\nZZZZZZ says 'goodbye'\n");	
                       #ifdef _WIN32
                       //  exit(EXIT_SUCCESS);
                       ExitProcess(0);
                       #else
                       exit(EXIT_SUCCESS);
                       #endif
   	                   }}
   	                   
   	                                                           
                                              
                       if (!strcmp(move, "force")) {
                       my_colour = NO_COLOUR;
                       force_mode = true;
   	                  // continue;
                       }
                                           
                       
                       if (!strcmp(move, "new")) {
                       FILE *f;
                       setbuf(stdout, NULL);
                       setbuf(stdin, NULL);
                       setvbuf(stdout, NULL, _IONBF, 0);
                       setvbuf(stdin, NULL, _IONBF, 0); 
                       BUG((f = fopen("game.txt", "w")) == NULL)
                       FCLOSE(f); 
                       my_colour = BLACK_COLOUR;
                       force_mode = false; 
                       DepthMax = max_depth; 
                       fen_to_board(start_fen, true);    
                       play_game(BLACK_COLOUR, WHITE_COLOUR, "game.txt");       
                       continue;
                       }
                       
                        
                       if (!strcmp(move, "protover")) {
                       printf("feature colors=1\n");
                       printf("feature myname=\"ZZZZZZ v%s, G. Wiesenekker\"\n",VERSION);
                       printf("feature time=1\n");
                       printf("feature pause=0\n");
                       printf("feature variants=\"normal\"\n");
                       printf("feature sigint=0\n");
                       printf("feature sigterm=0\n");
                       printf("feature ping=0\n");
                       printf("feature name=1\n");
                       printf("feature reuse=0\n");
                       printf("feature san=1\n");
                       printf("feature setboard=1\n");
                       printf("feature analyze=1\n");
                       printf("feature done=1\n");
   	                   continue;
                       }
         
                       if(!strcmp(move,"ping")) {
			           int ping;
			           sscanf(line,"%*s %d",&ping);
			           printf("pong %d\n",ping); 
			           continue;
		               }
         
                       if (!strcmp(move,"variant")) {
	    	           char variant[100];
	    	           scanf("%s",variant);
	    	           if (strcmp(variant,"normal")) {
			           printf("Error: (unsupported variant)\n");
	    	           }}    
          
                       if (!strcmp(move, "white")) {
                       opponent = true;
                       your_colour = BLACK_COLOUR;
                       my_colour = WHITE_COLOUR;               
                       continue;
                       }

                       if (!strcmp(move, "black")) {
                       opponent = true;
                       your_colour = WHITE_COLOUR;
                       my_colour = BLACK_COLOUR;     
                       continue;
                       }

                       if (!strcmp(move, "st")) {
                       sscanf(line, "st %d", &max_time);
			           max_time *= 1000;
			           stop_time = max_time/1000;
			           DepthMax = 10000;
			           continue;
		               }
          
                       if (!strcmp(move, "sd")) {
                       sscanf(line, "sd %d", &max_depth);
		               DepthMax = max_depth;
			           continue;
		               }
          
                       if (!strcmp(move, "time")) {       
                       sscanf(line, "time %d", &max_time);
                       DepthMax = 32;
                       max_time *= MAXTIME;  
                       max_time /= 400;
                       stop_time = max_time /= 400;
                      // if (max_time  < 18000) DepthMax = 10;  
                       continue;        
                       }
            
                       if (!strcmp(move, "level")) {       
                       sscanf(line, "level %d %d:%d", &max_moves,&max_time,&inc);
                       if (!max_moves) max_moves = 40;
                       if (!max_time)  max_time  =  1;
                       continue;        
                       }
                       
                       
                       if (!strcmp(move,"setboard")) {
                       sscanf(line, "setboard %s %s %s %s %s %s %s %s %s %s", buf, buf2, 
                       buf3, buf4, buf5, buf6, buf7, buf8, buf9, buf10, buf11, buf12);			
                       strcat(buf, " "), strcat(buf, buf2), strcat(buf, " "),
			           strcat(buf, buf3), strcat(buf, " "), strcat(buf, buf4),
			           strcat(buf, " "), strcat(buf, buf5), strcat(buf, " "),
			           strcat(buf, buf6),  strcat(buf, " ");
			           strcat(buf, " "), strcat(buf, buf7), strcat(buf, " "),
			           strcat(buf, buf8),  strcat(buf, " ");
			           strcat(buf, " "), strcat(buf, buf9), strcat(buf, " "),
			           strcat(buf, buf10),  strcat(buf, " ");
			           strcat(buf, " "), strcat(buf, buf11), strcat(buf, " "),
			           strcat(buf, buf12),  strcat(buf, " ");
                       strcpy(fen,buf);
                       fen_to_board(fen, true);
                       if (analyze_mode && !opponent) {
                       if (white_on_move) {
                       FILE *f;
                       BUG((f = fopen("game.txt", "w")) == NULL)
                       FCLOSE(f);                    
                       play_game(WHITE_COLOUR, BLACK_COLOUR, "game.txt");   
                       }else
                       if (black_on_move) {
                       FILE *f;
                       BUG((f = fopen("game.txt", "w")) == NULL)
                       FCLOSE(f);                    
                       play_game(BLACK_COLOUR, WHITE_COLOUR, "game.txt"); 
                       }
                       }      
                       using_setboard = true;
                       continue;
                	   }  
                       
                       if (!strcmp(move, "analyze")) {                
                       int iroot = 0;
                       max_time = 1000000;
                       stop_time = max_time;
                       analyze_on = true;
                       analyze_mode = true;
                       opponent = false;
                       DepthMax = 32;           
                       force_mode = true;  
                       gen_root();
                       if (nroot == 0){
                       DepthMax = 32;  
                       do_mate();                
                       }
                       DepthMax = 32;      
                       do_root(); 
                      }    		               		               
          
                       if (!strcmp(move, "result")) {
                       char result[256];
                       sscanf(line,"result %s",result);
	                   force_mode = true;
                       continue;
                       }
  	    
                       if (!strcmp(move, "xboard")) {
                       printf("feature done=0\n");
                       continue;          
                       }         
   
                                
        
if (!strcmp(move, "name")) {         
     continue;  
}                        
                    
if (!strcmp(move, "otim")) {
continue;
		}                  
if (!strcmp(move, "post")) {
  continue;
		}
if (!strcmp(move, "nopost")) {
          continue;
		}
if (!strcmp(move, ".")) {
         continue;
		}
if (!strcmp(move, "random")) {
         continue;
		}	
if (!strcmp(move, "hard")) {
         continue;
		}		      	
if (!strcmp(move, "easy")) {
         continue;
		}		      
if (!strcmp(move, "computer")) {
         continue;
		}		  
if (!strcmp(move, "accepted")) {
         continue;
		}		          
if (!strcmp(move, "rejected")) {
         continue;
		}	                 
                        
            iroot = search_root(move, MOVE_STRING_ANY, false, false);
            if (iroot == INVALID)
            continue;
            BUG((f = fopen(name, "a")) == NULL);
            move_to_san_string(root[iroot].root_move, true);
            fprintf(f, "%7s\n", move_string);
            FCLOSE(f);
                           
          continue;   
            }}}



///////////////////////////// main ///////////////////////////

int  CDECL
           main(int argc, char **argv)


{
 
  FILE *f;
  
  my_colour = BLACK_COLOUR;
  
  setbuf(stdout, NULL);
  setbuf(stdin, NULL);
  setvbuf(stdout, NULL, _IONBF, 0);
  setvbuf(stdin, NULL, _IONBF, 0); 
   
  signal(SIGINT, catch_sigint);

    my_printf("\nZZZZZZ version %s, Copyright (C) 1993-2009 G. Wiesenekker",VERSION);
    my_printf("\nWinboard protocol support by Jim Ablett %s\n\n",BUILDDATE);
    my_printf("ZZZZZZ comes with ABSOLUTELY NO WARRANTY.\n");
    my_printf("ZZZZZZ is free software, and you are welcome to redistribute it\n");
    my_printf("under certain conditions.\n");
    my_printf("See the file COPYING (of which you should have received a copy\n");
    my_printf("along with ZZZZZZ) for details.\n\n");
    
    initialize();

    
    printf("\nZZZZZZ runs only in a winboard compatible chess gui.\n");
    printf("Type 'quit' or 'exit' to terminate program.\n");
    printf("\n> ");
    
        
                   
       BUG((f = fopen("game.txt", "w")) == NULL)
     
                    FCLOSE(f);
        
        for (;;) {           
        play_game(BLACK_COLOUR, WHITE_COLOUR, "game.txt");       
       }
  
    return (EXIT_SUCCESS);
}

/*
   my_colour and your_colour are toggled by do_move and undo_move
 */

static int             my_colour,
                your_colour;

/*
   I know about cylinder boards, but they give me a headache!
   I use a 12x12 board, (8x8 with a double boundary).
 */



enum {
    A8 = 26, B8, C8, D8, E8, F8, G8, H8
};
enum {
    A7 = A8 + 12, B7, C7, D7, E7, F7, G7, H7
};
enum {
    A6 = A7 + 12, B6, C6, D6, E6, F6, G6, H6
};
enum {
    A5 = A6 + 12, B5, C5, D5, E5, F5, G5, H5
};
enum {
    A4 = A5 + 12, B4, C4, D4, E4, F4, G4, H4
};
enum {
    A3 = A4 + 12, B3, C3, D3, E3, F3, G3, H3
};
enum {
    A2 = A3 + 12, B2, C2, D2, E2, F2, G2, H2
};
enum {
    A1 = A2 + 12, B1, C1, D1, E1, F1, G1, H1
};

enum {
    a8 = 0, b8, c8, d8, e8, f8, g8, h8
};
enum {
    a7 = a8 + 8, b7, c7, d7, e7, f7, g7, h7
};
enum {
    a6 = a7 + 8, b6, c6, d6, e6, f6, g6, h6
};
enum {
    a5 = a6 + 8, b5, c5, d5, e5, f5, g5, h5
};
enum {
    a4 = a5 + 8, b4, c4, d4, e4, f4, g4, h4
};
enum {
    a3 = a4 + 8, b3, c3, d3, e3, f3, g3, h3
};
enum {
    a2 = a3 + 8, b2, c2, d2, e2, f2, g2, h2
};
enum {
    a1 = a2 + 8, b1, c1, d1, e1, f1, g1, h1
};

/*
   Sometimes we have to convert from the 'normal'
   index (0..63) to the 'true' index.
 */

static int       map[64] =
{
    A8, B8, C8, D8, E8, F8, G8, H8,
    A7, B7, C7, D7, E7, F7, G7, H7,
    A6, B6, C6, D6, E6, F6, G6, H6,
    A5, B5, C5, D5, E5, F5, G5, H5,
    A4, B4, C4, D4, E4, F4, G4, H4,
    A3, B3, C3, D3, E3, F3, G3, H3,
    A2, B2, C2, D2, E2, F2, G2, H2,
    A1, B1, C1, D1, E1, F1, G1, H1
};

/*
   mirror_map comes in handy when we have to initialize
   tables which are different for white and black.
 */

static int       mirror_map[64] =
{
    A1, B1, C1, D1, E1, F1, G1, H1,
    A2, B2, C2, D2, E2, F2, G2, H2,
    A3, B3, C3, D3, E3, F3, G3, H3,
    A4, B4, C4, D4, E4, F4, G4, H4,
    A5, B5, C5, D5, E5, F5, G5, H5,
    A6, B6, C6, D6, E6, F6, G6, H6,
    A7, B7, C7, D7, E7, F7, G7, H7,
    A8, B8, C8, D8, E8, F8, G8, H8
};

/*
   This needs no comment
 */

static char     *nota_table[64] =
{
    "a8", "b8", "c8", "d8", "e8", "f8", "g8", "h8",
    "a7", "b7", "c7", "d7", "e7", "f7", "g7", "h7",
    "a6", "b6", "c6", "d6", "e6", "f6", "g6", "h6",
    "a5", "b5", "c5", "d5", "e5", "f5", "g5", "h5",
    "a4", "b4", "c4", "d4", "e4", "f4", "g4", "h4",
    "a3", "b3", "c3", "d3", "e3", "f3", "g3", "h3",
    "a2", "b2", "c2", "d2", "e2", "f2", "g2", "h2",
    "a1", "b1", "c1", "d1", "e1", "f1", "g1", "h1"
};
static char     *nota_inval = "??";
static char    **nota;



/*
   The row numbers used by the program.
 */

static byte_t    row_table[64] =
{
    7, 7, 7, 7, 7, 7, 7, 7,
    6, 6, 6, 6, 6, 6, 6, 6,
    5, 5, 5, 5, 5, 5, 5, 5,
    4, 4, 4, 4, 4, 4, 4, 4,
    3, 3, 3, 3, 3, 3, 3, 3,
    2, 2, 2, 2, 2, 2, 2, 2,
    1, 1, 1, 1, 1, 1, 1, 1,
    0, 0, 0, 0, 0, 0, 0, 0
};
static int      *row;
static char     *nota_row = "12345678";


/*
   The column numbers used by the program
 */

static byte_t    column_table[64] =
{
    0, 1, 2, 3, 4, 5, 6, 7,
    0, 1, 2, 3, 4, 5, 6, 7,
    0, 1, 2, 3, 4, 5, 6, 7,
    0, 1, 2, 3, 4, 5, 6, 7,
    0, 1, 2, 3, 4, 5, 6, 7,
    0, 1, 2, 3, 4, 5, 6, 7,
    0, 1, 2, 3, 4, 5, 6, 7,
    0, 1, 2, 3, 4, 5, 6, 7
};
static int      *column;
static char     *nota_column = "abcdefgh";


//static char     *nota_colour = "wb.x";
static char           *colour_as_string[2] =
{"white ", "black "};

enum {
    PAWN = 0, KNIGHT, BISHOP, ROOK, QUEEN, KING
};

static char     *nota_kind = "PNBRQK";

/*
   To detect double pawns, free pawns, strong pieces etc. we need
   information about the pawn structure. We store the pawn structure
   in 8 bytes (one for each column) and if a square on a column
   is occupied by a pawn, the corresponding row bit in the byte for
   that column is set. The following tables are initialized at startup
   bit_count: gives the number of bits set in a byte.
   bit_low: gives the lowest bit set in a byte.
   bit_high: gives the highest bit set in a byte.
 */



static int      *bit_count;
static int      *bit_low;
static int      *bit_high;
static int       white_occup[8],
                black_occup[8];
static int      *occup[2] =
{white_occup, black_occup};


static int       value[6] =
{
    PAWN_SCORE, KNIGHT_SCORE, BISHOP_SCORE, ROOK_SCORE, QUEEN_SCORE, KING_SCORE
};

/*
   The king gradient is used in simple endgames. It forces the king to the
   side of the board.
 */

static byte_t    king_alone_grad_table[64] =
{
    -64, -32, -16, -16, -16, -16, -32, -64,
    -32, -16, -8, -8, -8, -8, -16, -32,
    -16, -8, -4, -4, -4, -4, -8, -16,
    -16, -8, -4, -2, -2, -4, -8, -16,
    -16, -8, -4, -2, -2, -4, -8, -16,
    -16, -8, -4, -4, -4, -4, -8, -16,
    -32, -16, -8, -8, -8, -8, -16, -32,
    -64, -32, -16, -16, -16, -16, -32, -64
};
static int      *king_alone_grad;

/*
   The pawn gradient urges pawns to promote.
 */

static byte_t    free_pawn_grad_table[64] =
{
    0, 0, 0, 0, 0, 0, 0, 0,
    64, 96, 96, 96, 96, 96, 96, 64,
    32, 48, 48, 48, 48, 48, 48, 32,
    16, 24, 24, 24, 24, 24, 24, 16,
    8, 12, 12, 12, 12, 12, 12, 8,
    4, 6, 6, 6, 6, 6, 6, 4,
    2, 3, 3, 3, 3, 3, 3, 2,
    0, 0, 0, 0, 0, 0, 0, 0
};
static int      *white_free_pawn_grad,
               *black_free_pawn_grad;
static int      *free_pawn_grad[2];

/*
   Visit tables for pawn and knight (only some fields are used in this version).
 */

static byte_t    pawn_visit_table[64] =
{
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 3, 3, 0, 0, 0,
    0, 0, 1, 2, 2, 1, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0
};

static byte_t    knight_visit_table[64] =
{
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 3, 0, 0, 3, 0, 0,
    0, 0, 0, 2, 2, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0
};

static int      *white_pawn_visit,
               *black_pawn_visit;
static int      *pawn_visit[2];

static int      *white_knight_visit,
               *black_knight_visit;
static int      *knight_visit[2];

static int      *pawn_board;

/*
   rooks on 7-th rank are strong.
 */

static int       strong_row[2] =
{6, 1};

/*
   attack_zero is used to clear the attack array.
 */

static int       attack_zero[BOARD_MAX] =
{
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1,
    -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1,
    -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1,
    -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1,
    -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1,
    -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1,
    -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1,
    -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1,
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1
};

static int      *attack_board;

/*
   movegen stuff
 */



/*
   We generate moves in a smart order: pawn captures first, then
   knight, bishop, rook, queen and king captures. Then en-passant and castling moves
   and finally non-capturing moves. When we search for captures, we store the
   captures and the special moves in a capture stack, and the normal moves
   in a move stack. move_top and capt_top are pointers to the current top of the
   move and capture stacks.
 */

static move_t   *move_stack,
               *capt_stack;
static move_t   *move_top;
move_t         *capt_top;

static int       shift[6] =
{false, false, true, true, true, false};

/*
   The offsets to generate the moves. We use the forward directions first.
   This may give more cutoffs (the moves are generated in
   a more natural order) but I never checked how much.
   The end of the list is denoted by offset 0. So you will often see
   loops like

   board_t *p;
   int *d;
   for (p = list[icolour][ikind]; p->board_kind != INVALID; p++)
   {
   if (p->board_where == INVALID) continue;
   for (d = p->board_dir; *d != 0; d++)
   {
   ..
   }
   }

 */

static int       white_pawn_dir[4] =
{-12, -13, -11, 0};
static int       black_pawn_dir[4] =
{12, 11, 13, 0};
static int      *pawn_dir[2] =
{white_pawn_dir, black_pawn_dir};

static int       white_knight_dir[9] =
{-25, -23, -14, -10, 10, 14, 23, 25, 0};
static int       black_knight_dir[9] =
{25, 23, 14, 10, -10, -14, -23, -25, 0};
static int      *knight_dir[2] =
{white_knight_dir, black_knight_dir};

static int       white_bishop_dir[5] =
{-13, -11, 11, 13, 0};
static int       black_bishop_dir[5] =
{13, 11, -11, -13, 0};
static int      *bishop_dir[2] =
{white_bishop_dir, black_bishop_dir};

static int       white_rook_dir[5] =
{-12, -1, 1, 12, 0};
static int       black_rook_dir[5] =
{12, 1, -1, -12, 0};
static int      *rook_dir[2] =
{white_rook_dir, black_rook_dir};

static int       white_queen_dir[9] =
{-13, -12, -11, -1, 1, 11, 12, 13, 0};
static int       black_queen_dir[9] =
{13, 12, 11, 1, -1, -11, -12, -13, 0};

static int       white_king_dir[9] =
{-13, -12, -11, -1, 1, 11, 12, 13, 0};
static int       black_king_dir[9] =
{13, 12, 11, 1, -1, -11, -12, -13, 0};

static int      *white_dir[6] =
{white_pawn_dir, white_knight_dir, white_bishop_dir, white_rook_dir, white_queen_dir, white_king_dir};
static int      *black_dir[6] =
{black_pawn_dir, black_knight_dir, black_bishop_dir, black_rook_dir, black_queen_dir, black_king_dir};
static int     **dir[2] =
{white_dir, black_dir};

/*
   A pawn promotes if it is on row 7 for white or row 0 for black.
 */

static int       prom_row[2] =
{7, 0};

/*
   hack_ep keeps track of possible en passant right (will be INVALID
   most of the time.
 */

static int       hack_ep;

/*
   We store the pieces in lists, one list for every piece.
   The board is an array of pointers to these lists. Two special
   list entries are the empty square and the side square.
 */

static board_t  *board[BOARD_MAX];
static board_t   empty =
{NONE, INVALID, 0, INVALID, true, false, NULL, false, NULL};
static board_t   side =
{SIDE, INVALID, 0, INVALID, true, false, NULL, false, NULL};

static board_t   white_pawn_list[LIST_MAX],
                black_pawn_list[LIST_MAX];
static board_t  *pawn_list[2] =
{white_pawn_list, black_pawn_list};

static board_t   white_knight_list[LIST_MAX],
                black_knight_list[LIST_MAX];
static board_t  *knight_list[2] =
{white_knight_list, black_knight_list};

static board_t   white_bishop_list[LIST_MAX],
                black_bishop_list[LIST_MAX];
static board_t  *bishop_list[2] =
{white_bishop_list, black_bishop_list};

static board_t   white_rook_list[LIST_MAX],
                black_rook_list[LIST_MAX];
static board_t  *rook_list[2] =
{white_rook_list, black_rook_list};

static board_t   white_queen_list[LIST_MAX],
                black_queen_list[LIST_MAX];
static board_t  *queen_list[2] =
{white_queen_list, black_queen_list};

static board_t   white_king_list[LIST_MAX],
                black_king_list[LIST_MAX];
static board_t  *king_list[2] =
{white_king_list, black_king_list};

static board_t  *white_list[6] =
{white_pawn_list, white_knight_list, white_bishop_list, white_rook_list, white_queen_list, white_king_list};
static board_t  *black_list[6] =
{black_pawn_list, black_knight_list, black_bishop_list, black_rook_list, black_queen_list, black_king_list};
static board_t **list[2] =
{white_list, black_list};

static int       white_piece_count[6];
static int       black_piece_count[6];
static int      *piece_count[2] =
{white_piece_count, black_piece_count};
static int             tot_count[2];



static var_t    *var_stack;





static node_t   *node_stack;
node_t         *node;
static int       inode,
                inode_root,
                inode_max;




/*
   the pointers to the tables with hash keys.
 */


static hash_key_t *hash_key[2][6];
static hash_key_t *ep_key;
static hash_key_t *castle_key;


/*
   cache_best is a pointer to the best move (it is the number of the best move
   in the list of pseudo legal moves).
   cache_score is the score.
   cache_key is the key
 */

#define FEN_STRING_MAX 96

typedef struct {
    int             cache_best;
    score_t         cache_score;
    hash_key_t      cache_key;
#ifdef CACHE_FEN
    char            cache_fen[FEN_STRING_MAX];
#endif
}
                cache_t;

/*
   For the static evaluation hash table we do not have a best move.
 */

typedef struct {
    score_t         cache_score;
    hash_key_t      cache_key;
#ifdef CACHE_FEN
    char            cache_fen[FEN_STRING_MAX];
#endif
}
                score_cache_t;

/*
   We use separate hash tables for white and black
 */


static cache_t  *cache[2];
static score_cache_t *score_cache[2];


/*
   Some counters for statistics.
 */

static long      ngens,
                ndos;
static long      nevals,
                nhits;
#ifdef BUG_DEPTH
static int       bug_depth[NODE_MAX];
#endif

/*
   Variables for time control
 */

static clock_t   clock_begin,
                clock_end;
static double    time_used;

/*
   move_to_string converts the move to a long algebraic notation
   and stores the move in move_string.
 */

char            move_string[MOVE_STRING_MAX];

void            move_to_string(move_t m)
{
    long            kind;
    int             from,
                    to;
    char            c;

    UNPACK_MOVE(m, kind, from, to)
        if (kind == NULL_MOVE) {
        sprintf(move_string, "%s", "NUL");
    } else if (kind & PROM_CAPT) {
        if (kind & PROM_QUEEN)
            c = 'Q';
        else if (kind & PROM_ROOK)
            c = 'R';
        else if (kind & PROM_BISHOP)
            c = 'B';
        else
            c = 'N';
        sprintf(move_string, "%s%s%c", GET_NOTA(from), GET_NOTA(to), c);
    } else if (kind & PROM_NORM) {
        if (kind & PROM_QUEEN)
            c = 'Q';
        else if (kind & PROM_ROOK)
            c = 'R';
        else if (kind & PROM_BISHOP)
            c = 'B';
        else
            c = 'N';
        sprintf(move_string, "%s%s%c", GET_NOTA(from), GET_NOTA(to), c);
    } else if (kind == EP)
        sprintf(move_string, "%s%s", GET_NOTA(from), GET_NOTA(to));
    else if (kind == SHORT)
        sprintf(move_string, "o-o");
    else if (kind == LONG)
        sprintf(move_string, "o-o-o");
    else if (kind == CAPT)
        sprintf(move_string, "%s%s", GET_NOTA(from), GET_NOTA(to));
    else
        sprintf(move_string, "%s%s", GET_NOTA(from), GET_NOTA(to));
}



/*
   board_to_fen converts a board position to FEN notation
 */

static char     *white_fen[6] =
{"P", "N", "B", "R", "Q", "K"};
static char     *black_fen[6] =
{"p", "n", "b", "r", "q", "k"};
static char     *empty_fen[9] =
{"", "1", "2", "3", "4", "5", "6", "7", "8"};
static char     *file_fen[9] =
{"a", "b", "c", "d", "e", "f", "g", "h"};
//static char     *start_fen = "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1";
static char     *start_board = "rnbqkbnr/pppppppp/eeeeeeee/eeeeeeee/eeeeeeee/eeeeeeee/PPPPPPPP/RNBQKBNR";

/*
   We also use the fen string for the database. As we are then
   only interested in transpositions, we include the en-passant square
   only if an en-passant capture is possible, and we omit the half and
   full move fields.
 */

void            board_to_fen(char *string, int strict)
{
    int             i,
                    j,
                    k;
    int             n;
    board_t        *p;

    strcpy(string, "");

    k = 0;
    for (j = 0; j < 8; j++) {
        n = 0;
        for (i = 0; i < 8; i++) {
            p = board[map[k]];
            if (p == &empty) {
                n++;
            } else {
                strcat(string, empty_fen[n]);
                n = 0;
                if (p->board_colour == WHITE_COLOUR)
                    strcat(string, white_fen[p->board_kind]);
                else if (p->board_colour == BLACK_COLOUR)
                    strcat(string, black_fen[p->board_kind]);
                else
                    FATAL_ERROR("p->board_colour out of range");
            }
            k++;
        }
        strcat(string, empty_fen[n]);
        n = 0;
        if (j < 7)
            strcat(string, "/");
        else
            strcat(string, " ");
    }
    if (my_colour == WHITE_COLOUR)
        strcat(string, "w ");
    else if (my_colour == BLACK_COLOUR)
        strcat(string, "b ");
    else
        FATAL_ERROR("my_colour out of range");
    /* castling */

    n = 0;
    if (!board[E1]->board_moved) {
        if (!board[H1]->board_moved) {
            strcat(string, "K");
            n++;
        }
        if (!board[A1]->board_moved) {
            strcat(string, "Q");
            n++;
        }
    }
    if (!board[E8]->board_moved) {
        if (!board[H8]->board_moved) {
            strcat(string, "k");
            n++;
        }
        if (!board[A1]->board_moved) {
            strcat(string, "q");
            n++;
        }
    }
    if (n == 0)
        strcat(string, "- ");
    else
        strcat(string, " ");

    /* en passant */

    if (hack_ep == INVALID)
        strcat(string, "-");
    else {
        int             ep;

        if (strict) {
            ep = true;
        } else {
            ep = false;
            j = hack_ep - 1;
            if ((board[j]->board_colour == my_colour) and(board[j]->board_kind == PAWN))
                ep = true;
            j = hack_ep + 1;
            if ((board[j]->board_colour == my_colour) and(board[j]->board_kind == PAWN))
                ep = true;
        }
        if (ep) {
            strcat(string, file_fen[column[hack_ep]]);
            if (row[hack_ep] == 3)
                strcat(string, "3");
            else if (row[hack_ep] == 4)
                strcat(string, "6");
            else
                FATAL_ERROR("hack_ep out of range");
        } else
            strcat(string, "-");
    }
    /* remaining fields are dummy for the moment */

    if (strict)
        strcat(string, " 0 0");
}

/*
   return_cur_key returns the current hash key.
 */

static           hash_key_t
                return_cur_key(void)
{
    int             icolour,
                    ikind;
    hash_key_t      key;
    board_t        *p;

    CLEAR_KEY(key);
    for (icolour = WHITE_COLOUR; icolour <= BLACK_COLOUR; icolour++) {
        for (ikind = PAWN; ikind <= KING; ikind++) {
            for (p = list[icolour][ikind]; p->board_kind != INVALID; p++) {
                if (p->board_where != INVALID)
                    UPDATE_KEY(key, p->board_key[p->board_where]);
            }
        }
    }
    return (key);
}



/*
   return_mat_score calculates the current material value
 */


static int       return_mat_score(int my_colour)
{
    int             sum;
    int             ikind;
    board_t        *p;

    sum = 0;
    for (ikind = PAWN; ikind <= QUEEN; ikind++) {
        for (p = list[my_colour][ikind]; p->board_kind != INVALID; p++)
            if (p->board_where != INVALID)
                sum += p->board_value;
    }
    return (sum);
}

/*
   update_key updates the hash key with ep rights and castling rights.
   This makes the position unique.
 */

static void      update_key(int your_colour, hash_key_t * key)
{
    int             j;

    if (hack_ep != INVALID) {
        /* Pawn to the left? */

        j = hack_ep - 1;
        if ((board[j]->board_colour == your_colour) and(board[j]->board_kind == PAWN))
            UPDATE_KEY(*key, ep_key[j]);
        /* Pawn to the right? */

        j = hack_ep + 1;
        if ((board[j]->board_colour == your_colour) and(board[j]->board_kind == PAWN))
            UPDATE_KEY(*key, ep_key[j]);
    }
    if (!board[E1]->board_moved) {
        if (!board[H1]->board_moved)
            UPDATE_KEY(*key, castle_key[H1]);
        if (!board[A1]->board_moved)
            UPDATE_KEY(*key, castle_key[A1]);
    }
    if (!board[E8]->board_moved) {
        if (!board[H8]->board_moved)
            UPDATE_KEY(*key, castle_key[H8]);
        if (!board[A1]->board_moved)
            UPDATE_KEY(*key, castle_key[A8]);
    }
}

/*
   check_lists checks whether the lists of pieces are still correct and returns
   an error code if there is something wrong. Returns 0 if the lists are OK.
 */


static int       check_list(void)
{
    int             i;
    int             icolour;
    board_t        *p;
    int             ncolour,
                    nkind;
    int             ikind;

    for (i = 0; i < 64; i++)
        if ((board[map[i]] != &empty) and(board[map[i]]->board_where != map[i]))
            return (1);
    for (icolour = WHITE_COLOUR; icolour <= BLACK_COLOUR; icolour++) {
        /* ncolour is the total number of pieces of colour icolour. */

        ncolour = 0;
        for (ikind = PAWN; ikind <= KING; ikind++) {
            /* nkind is the total number of pieces of kind ikind. */

            nkind = 0;
            for (p = list[icolour][ikind]; p->board_kind != INVALID; p++) {
                if (p->board_where != INVALID) {
                    if (board[p->board_where] != p)
                        return (2);
                    nkind++;
                }
            }
            if (nkind != piece_count[icolour][ikind])
                return (3);
            ncolour += nkind;
        }
        if (ncolour != tot_count[icolour])
            return (4);
    }
    return (0);
}

/*
   compact_lists removes captured pieces from the lists.
   Is called before we start thinking about a position.
 */

void            compact_lists(void)
{
    int             icolour,
                    ikind;
    board_t        *p;
    int             err;
    int             n,
                    i;
    board_t         temp_list[LIST_MAX];

    for (icolour = WHITE_COLOUR; icolour <= BLACK_COLOUR; icolour++) {
        /* We skip the king of course. */

        for (ikind = PAWN; ikind <= QUEEN; ikind++) {
            /* Copy all pieces to temp_list. */

            n = 0;
            for (p = list[icolour][ikind]; p->board_kind != INVALID; p++) {
                if (p->board_where != INVALID) {
                    BUG(n >= LIST_MAX)
                        temp_list[n] = *p;
                    n++;
                }
            }

            /* Copy back. */

            p = list[icolour][ikind];
            for (i = 0; i < n; i++) {
                *p = temp_list[i];
                board[p->board_where] = p;
                p++;
            }

            /* Remaining entries are invalid. */

            for (; i < LIST_MAX; i++) {
                p->board_kind = p->board_where = INVALID;
                p++;
            }
        }
    }
    if ((err = check_list()) > 0) {
        my_printf("err=%d\n", err);
      
        FATAL_ERROR("list error");
    }
}

/*
   fen_to_board sets the board
 */

void            fen_to_board(char *string, int strict)
{
    int             i,
                    j,
                    k,
                    n;
    int             icolour,
                    ikind;
    board_t        *p;
    int             err;
    char           *sync;

    hack_ep = INVALID;

    /* Initialize the boundary and the board. */

    for (i = 0; i < BOARD_MAX; i++)
        board[i] = &side;
    for (i = 0; i < 64; i++)
        board[map[i]] = &empty;

    for (icolour = WHITE_COLOUR; icolour <= BLACK_COLOUR; icolour++) {
        tot_count[icolour] = 0;
        for (ikind = PAWN; ikind <= KING; ikind++) {
            /* Initialize all entries. */

            p = list[icolour][ikind];
            for (i = 0; i < LIST_MAX; i++) {
                p->board_colour = icolour;
                p->board_kind = INVALID;
                p->board_value = value[ikind];
                p->board_where = INVALID;
                p->board_moved = true;
                p->board_captured = false;
                p->board_key = hash_key[icolour][ikind];
                p->board_shift = shift[ikind];
                p->board_dir = dir[icolour][ikind];
                p->board_count = piece_count[icolour] + ikind;
                p++;
            }
            piece_count[icolour][ikind] = 0;
        }
    }

    k = 0;
    sync = start_board;
    for (i = 0; i < 8; i++) {
        j = 0;
        while (j < 8) {
            for (n = 8; n > 0; n--) {
                if (*string == *empty_fen[n])
                    break;
            }
            if (n > 0) {
                j += n;
                k += n;
                sync += n;
                string++;
                continue;
            }
            icolour = INVALID;
            for (ikind = PAWN; ikind <= KING; ikind++) {
                if (*string == *black_fen[ikind]) {
                    icolour = BLACK_COLOUR;
                    break;
                }
                if (*string == *white_fen[ikind]) {
                    icolour = WHITE_COLOUR;
                    break;
                }
            }
            BUG(icolour == INVALID)
                p = list[icolour][ikind] + piece_count[icolour][ikind];
            p->board_kind = ikind;
            p->board_where = map[k];
            board[map[k]] = p;
            piece_count[icolour][ikind]++;
            tot_count[icolour]++;

            /* We assume a piece has not moved if it is still on it's
             * starting position. */

            if (*string == *sync)
                p->board_moved = false;
            else
                p->board_moved = true;
            j++;
            k++;
            string++;
            sync++;
        }
        if (i < 7) {
            BUG(*string != '/');
            BUG(*sync != '/');
        }
        string++;
        sync++;
    }

    if (*string == 'w')
	{
        WHITE_TO_MOVE;
        white_on_move = true;
        black_on_move = false;
     
	}
    else if (*string == 'b')
	{
        BLACK_TO_MOVE;
        black_on_move = true;
        white_on_move = false;
      
	}
    else
        FATAL_ERROR("colour out of range!");
    string++;

    BUG(*string != ' ');
    string++;

    /* possible castling rights */

    if (*string == '-') {
        board[E1]->board_moved = true;
        board[A1]->board_moved = true;
        board[H1]->board_moved = true;
        board[E8]->board_moved = true;
        board[A8]->board_moved = true;
        board[H8]->board_moved = true;
        string++;
    } else if (*string == '?') {
        if (strict)
            FATAL_ERROR("invalid castle string");
        string++;
    } else {
        while (*string != ' ') {
            if (*string == 'K') {
                board[E1]->board_moved = false;
                board[H1]->board_moved = false;
            } else if (*string == 'Q') {
                board[E1]->board_moved = false;
                board[A1]->board_moved = false;
            } else if (*string == 'k') {
                board[E8]->board_moved = false;
                board[H8]->board_moved = false;
            } else if (*string == 'q') {
                board[E8]->board_moved = false;
                board[A8]->board_moved = false;
            } else
                FATAL_ERROR("fen error");
            string++;
        }
    }

    BUG(*string != ' ')
        string++;

    if (*string == '-') {
        hack_ep = INVALID;
        string++;
    } else {
        for (k = 0; k < 64; k++) {
            if (strncmp(string, nota_table[k], 2) == 0) {
                hack_ep = map[k] + pawn_dir[your_colour][0];
                break;
            }
        }
        BUG(k == 64)
            string++;
        string++;
    }

    if (strict) {
        BUG(*string != ' ')
            string++;
    }
    BUG(piece_count[WHITE_COLOUR][KING] != 1)
        BUG(piece_count[BLACK_COLOUR][KING] != 1)
        capt_top = capt_stack;

    /* Set node. */

    node = node_stack;
    node->node_key = return_cur_key();
    node->node_total_key = node->node_key;
    update_key(your_colour, &(node->node_total_key));
    node->node_mat[WHITE_COLOUR] = return_mat_score(WHITE_COLOUR);
    node->node_mat[BLACK_COLOUR] = return_mat_score(BLACK_COLOUR);
    node->node_reverse = false;
    node->node_hack_ep = INVALID;

    if ((err = check_list()) > 0) {
        my_printf("err=%d\n", err);
    
        FATAL_ERROR("list error");
    }
}

/*
   attacked returns true if square i is attacked by your_colour.
 */

static int       attacked(int i, int your_colour)
{
    int            *d;
    board_t        *p;
    int             j;

    p = board[i - pawn_dir[your_colour][1]];
    if ((p->board_colour == your_colour) and(p->board_kind == PAWN))
        return (true);
    p = board[i - pawn_dir[your_colour][2]];
    if ((p->board_colour == your_colour) and(p->board_kind == PAWN))
        return (true);
    for (d = knight_dir[your_colour]; *d != 0; d++) {
        p = board[i - *d];
        if ((p->board_colour == your_colour) and(p->board_kind == KNIGHT))
            return (true);
    }
    for (d = bishop_dir[your_colour]; *d != 0; d++) {
        j = i - *d;
        p = board[j];
        if ((p->board_colour == your_colour) and(p->board_kind == KING))
            return (true);
        while (board[j] == &empty)
            j -= *d;
        p = board[j];
        if ((p->board_colour == your_colour) and
            ((p->board_kind == BISHOP) or(p->board_kind == QUEEN)))
            return (true);
    }
    for (d = rook_dir[your_colour]; *d != 0; d++) {
        j = i - *d;
        p = board[j];
        if ((p->board_colour == your_colour) and(p->board_kind == KING))
            return (true);
        while (board[j] == &empty)
            j -= *d;
        p = board[j];
        if ((p->board_colour == your_colour) and
            ((p->board_kind == ROOK) or(p->board_kind == QUEEN)))
            return (true);
    }
    return (false);
}

int             i_am_check(int my_colour, int your_colour)
{
    return (attacked(king_list[my_colour]->board_where, your_colour));
}

/*
   gen_moves generates the moves for the piece on square p.
 */

static void      gen_moves(int my_colour, int your_colour, board_t * p)
{
    int             j;
    int            *d;
    int             irow;

    if (p->board_kind == PAWN) {
        /* Pawn captures. */

        j = p->board_where + p->board_dir[1];
        irow = row[j];
        if (board[j]->board_colour == your_colour) {
            if (irow == prom_row[my_colour]) {
                ADD_MOVE(capt_top, (PROM_CAPT | PROM_QUEEN), p->board_where, j)
                    ADD_MOVE(capt_top, (PROM_CAPT | PROM_ROOK), p->board_where, j)
                    ADD_MOVE(capt_top, (PROM_CAPT | PROM_BISHOP), p->board_where, j)
                    ADD_MOVE(capt_top, (PROM_CAPT | PROM_KNIGHT), p->board_where, j)
            } else {
                ADD_MOVE(capt_top, CAPT, p->board_where, j)
            }
        }
        j = p->board_where + p->board_dir[2];
        irow = row[j];
        if (board[j]->board_colour == your_colour) {
            if (irow == prom_row[my_colour]) {
                ADD_MOVE(capt_top, (PROM_CAPT | PROM_QUEEN), p->board_where, j)
                    ADD_MOVE(capt_top, (PROM_CAPT | PROM_ROOK), p->board_where, j)
                    ADD_MOVE(capt_top, (PROM_CAPT | PROM_BISHOP), p->board_where, j)
                    ADD_MOVE(capt_top, (PROM_CAPT | PROM_KNIGHT), p->board_where, j)
            } else {
                ADD_MOVE(capt_top, CAPT, p->board_where, j)
            }
        }
        /* non-capturing moves for pawn. */

        j = p->board_where + p->board_dir[0];
        if (board[j] == &empty) {
            irow = row[j];
            if (irow == prom_row[my_colour]) {
                ADD_MOVE(move_top, (PROM_NORM | PROM_QUEEN), p->board_where, j)
                    ADD_MOVE(move_top, (PROM_NORM | PROM_ROOK), p->board_where, j)
                    ADD_MOVE(move_top, (PROM_NORM | PROM_BISHOP), p->board_where, j)
                    ADD_MOVE(move_top, (PROM_NORM | PROM_KNIGHT), p->board_where, j)
            } else {
                ADD_MOVE(move_top, NORM, p->board_where, j)
            }

            /* double pawn move */

            if (!p->board_moved) {
                j += p->board_dir[0];
                if (board[j] == &empty) {
                    ADD_MOVE(move_top, DOUBLE, p->board_where, j)
                }
            }
        }
    } else {
        /* Other pieces. */

        for (d = p->board_dir; *d != 0; d++) {
            j = p->board_where;
            while (true) {
                /* Try next square. */

                j += *d;
                if (board[j] == &empty) {
                    ADD_MOVE(move_top, NORM, p->board_where, j)
                        if (!p->board_shift)
                        break;
                } else {
                    /* May be capture . */

                    if (board[j]->board_colour == your_colour) {
                        ADD_MOVE(capt_top, CAPT, p->board_where, j)
                    }
                    break;
                }
            }
        }
    }
}

/*
   gen_special_moves generates en passant and short or long castle.
 */

static void      gen_special_moves(int my_colour, int your_colour)
{
    int             j;

    if (hack_ep != INVALID) {
        /* Pawn to the left? */

        j = hack_ep - 1;
        if ((board[j]->board_colour == my_colour) and(board[j]->board_kind == PAWN)) {
            ADD_MOVE(capt_top, EP, j, hack_ep + pawn_dir[my_colour][0])
        }
        /* Pawn to the right? */

        j = hack_ep + 1;
        if ((board[j]->board_colour == my_colour) and(board[j]->board_kind == PAWN)) {
            ADD_MOVE(capt_top, EP, j, hack_ep + pawn_dir[my_colour][0])
        }
    }
    if (my_colour == WHITE_COLOUR) {
        /* empty square has board_moved set to true. Dirty trick eh? */

        if (!board[E1]->board_moved) {
            if (!board[H1]->board_moved) {
                if ((board[F1] == &empty) and(board[G1] == &empty) and
                    ! attacked(E1, your_colour) and ! attacked(F1, your_colour) and ! attacked(G1, your_colour)) {
                    ADD_MOVE(capt_top, SHORT, 0, 0)
                }
            }
            if (!board[A1]->board_moved) {
                if ((board[D1] == &empty) and(board[C1] == &empty) and
                    (board[B1] == &empty) and
                    ! attacked(E1, your_colour) and ! attacked(D1, your_colour) and ! attacked(C1, your_colour)) {
                    ADD_MOVE(capt_top, LONG, 0, 0)
                }
            }
        }
    } else {
        if (!board[E8]->board_moved) {
            if (!board[H8]->board_moved) {
                if ((board[F8] == &empty) and(board[G8] == &empty) and
                    ! attacked(E8, your_colour) and ! attacked(F8, your_colour) and ! attacked(G8, your_colour)) {
                    ADD_MOVE(capt_top, SHORT, 0, 0)
                }
            }
            if (!board[A8]->board_moved) {
                if ((board[D8] == &empty) and(board[C8] == &empty) and
                    (board[B8] == &empty) and
                    ! attacked(E8, your_colour) and ! attacked(D8, your_colour) and ! attacked(C8, your_colour)) {
                    ADD_MOVE(capt_top, LONG, 0, 0)
                }
            }
        }
    }
}

/*
   gen_all_moves generates all moves for a colour.
 */

int             gen_all_moves(void)
{
    int             ikind,
                    imove;
    board_t        *p;
    int             nmoves;
    move_t         *base;

    move_top = move_stack;
    base = capt_top;
    for (ikind = PAWN; ikind <= KING; ikind++) {
        for (p = list[my_colour][ikind]; p->board_kind != INVALID; p++) {
            if (p->board_where == INVALID)
                continue;
            gen_moves(my_colour, your_colour, p);
        }
    }
    gen_special_moves(my_colour, your_colour);
    nmoves = move_top - move_stack;
    for (imove = 0; imove < nmoves; imove++)
        *capt_top++ = move_stack[imove];
    nmoves = capt_top - base;
#if DEBUG_LEVEL > 0
    BUG((capt_top - capt_stack) >= CAPT_MAX)
#endif
        return (nmoves);
}

/*
   gen_capts generates all captures to a square. Same as attacked routine,
   except ADD_MOVE is added.
 */

static int       gen_capts(int i, int your_colour)
{
    int            *d;
    board_t        *p;
    int             j;
    move_t         *base;

    base = capt_top;
    j = i - pawn_dir[your_colour][1];
    p = board[j];
    if ((p->board_colour == your_colour) and(p->board_kind == PAWN)) {
        if (row[i] == prom_row[your_colour]) {
            ADD_MOVE(capt_top, (PROM_CAPT | PROM_QUEEN), j, i);
            ADD_MOVE(capt_top, (PROM_CAPT | PROM_ROOK), j, i);
            ADD_MOVE(capt_top, (PROM_CAPT | PROM_BISHOP), j, i);
            ADD_MOVE(capt_top, (PROM_CAPT | PROM_KNIGHT), j, i);
        } else
            ADD_MOVE(capt_top, CAPT, j, i);
    }
    j = i - pawn_dir[your_colour][2];
    p = board[j];
    if ((p->board_colour == your_colour) and(p->board_kind == PAWN)) {
        if (row[i] == prom_row[your_colour]) {
            ADD_MOVE(capt_top, (PROM_CAPT | PROM_QUEEN), j, i);
            ADD_MOVE(capt_top, (PROM_CAPT | PROM_ROOK), j, i);
            ADD_MOVE(capt_top, (PROM_CAPT | PROM_BISHOP), j, i);
            ADD_MOVE(capt_top, (PROM_CAPT | PROM_KNIGHT), j, i);
        } else
            ADD_MOVE(capt_top, CAPT, j, i);
    }
    for (d = knight_dir[your_colour]; *d != 0; d++) {
        j = i - *d;
        p = board[j];
        if ((p->board_colour == your_colour) and(p->board_kind == KNIGHT))
            ADD_MOVE(capt_top, CAPT, j, i);
    }
    for (d = bishop_dir[your_colour]; *d != 0; d++) {
        j = i - *d;
        p = board[j];
        if ((p->board_colour == your_colour) and(p->board_kind == KING))
            ADD_MOVE(capt_top, CAPT, j, i);
        while (board[j] == &empty)
            j -= *d;
        p = board[j];
        if ((p->board_colour == your_colour) and
            ((p->board_kind == BISHOP) or(p->board_kind == QUEEN)))
            ADD_MOVE(capt_top, CAPT, j, i);
    }
    for (d = rook_dir[your_colour]; *d != 0; d++) {
        j = i - *d;
        p = board[j];
        if ((p->board_colour == your_colour) and(p->board_kind == KING))
            ADD_MOVE(capt_top, CAPT, j, i);
        while (board[j] == &empty)
            j -= *d;
        p = board[j];
        if ((p->board_colour == your_colour) and
            ((p->board_kind == ROOK) or(p->board_kind == QUEEN)))
            ADD_MOVE(capt_top, CAPT, j, i);
    }
    return (capt_top - base);
}

/*
   Chess do_move routines always seem to get a bit messy when you add features
   like piece lists and piece counts..
 */

void            do_move(move_t m)
{
/*
   DO_CAPT removes a piece on SQUARE B.
 */

#define DO_CAPT(B)\
        {\
                p = board[B];\
                if (p->board_colour != your_colour) {my_printf("ep=%s %s\n", GET_NOTA(hack_ep), colour_as_string[my_colour]);}\
                BUG(p->board_kind == KING)\
                node->capt = p;\
                p->board_where = INVALID;\
                UPDATE_KEY(node->node_key, p->board_key[B])\
                node->node_mat[your_colour] -= p->board_value;\
                tot_count[your_colour]--;\
                (*p->board_count)--;\
        }

/*
   DO_MOVE moves a piece from square A to square B.
 */

#define DO_MOVE(A, B)\
        {\
                p = board[A];\
                UPDATE_KEY(node->node_key, p->board_key[A])\
                UPDATE_KEY(node->node_key, p->board_key[B])\
                p->board_where = B;\
                node->node_moved = p->board_moved;\
                node->node_captured = p->board_captured;\
                p->board_moved = true;\
                board[A] = &empty;\
                board[B] = p;\
        }

    board_t        *p;
    long            kind;
    int             from,
                    to;
    int             ep;
#if DEBUG_LEVEL > 0
    int             err;
    hash_key_t      temp_key;
#endif


    /* Save move and hack_ep for undo move */

    
    node->node_move = m;
    node->node_hack_ep = hack_ep;
    hack_ep = INVALID;

    /* Copy incrementally updated information to next node. */


    node[1].node_key = node->node_key;
    node[1].node_mat[WHITE_COLOUR] = node->node_mat[WHITE_COLOUR];
    node[1].node_mat[BLACK_COLOUR] = node->node_mat[BLACK_COLOUR];
    node++;
    inode++;

    node->node_reverse = false;
    UNPACK_MOVE(m, kind, from, to)
        if (kind == NULL_MOVE)
        goto label_null;


    /* Another cheap check. */


    if ((kind != SHORT) and(kind != LONG)) {
        BUG(column[from] == INVALID)
            BUG(column[to] == INVALID)
    }
    if (kind == EP) {
        ep = to - pawn_dir[my_colour][0];
        DO_CAPT(ep)
            board[ep] = &empty;
        DO_MOVE(from, to)
    } else if (kind & PROM_CAPT) {
        DO_CAPT(to)
        /* Remove pawn */


            p = board[from];
        node->prom = p;
        board[from] = &empty;
        p->board_where = INVALID;
        (*p->board_count)--;
        UPDATE_KEY(node->node_key, p->board_key[from]);
        node->node_mat[my_colour] -= p->board_value;


        /* Add piece. Search end of list first. */


        if (kind & PROM_QUEEN) {
            for (p = queen_list[my_colour]; p->board_kind == QUEEN; p++);
            p->board_kind = QUEEN;
        } else if (kind & PROM_ROOK) {
            for (p = rook_list[my_colour]; p->board_kind == ROOK; p++);
            p->board_kind = ROOK;
        } else if (kind & PROM_BISHOP) {
            for (p = bishop_list[my_colour]; p->board_kind == BISHOP; p++);
            p->board_kind = BISHOP;
        } else {
            for (p = knight_list[my_colour]; p->board_kind == KNIGHT; p++);
            p->board_kind = KNIGHT;
        }
        p->board_where = to;
        (*p->board_count)++;
        board[to] = p;
        UPDATE_KEY(node->node_key, p->board_key[to]);
        node->node_mat[my_colour] += p->board_value;
    } else if (kind == SHORT) {
        if (my_colour == WHITE_COLOUR) {
            DO_MOVE(E1, G1)
                DO_MOVE(H1, F1)
        } else {
            DO_MOVE(E8, G8)
                DO_MOVE(H8, F8)
        }
    } else if (kind == LONG) {
        if (my_colour == WHITE_COLOUR) {
            DO_MOVE(E1, C1)
                DO_MOVE(A1, D1)
        } else {
            DO_MOVE(E8, C8)
                DO_MOVE(A8, D8)
        }
    } else if (kind & PROM_NORM) {
        /* Remove pawn */
        p = board[from];
        node->prom = p;
        board[from] = &empty;
        p->board_where = INVALID;
        (*p->board_count)--;
        UPDATE_KEY(node->node_key, p->board_key[from]);
        node->node_mat[my_colour] -= p->board_value;

        /* Add piece */


        if (kind & PROM_QUEEN) {
            for (p = queen_list[my_colour]; p->board_kind == QUEEN; p++);
            p->board_kind = QUEEN;
        } else if (kind & PROM_ROOK) {
            for (p = rook_list[my_colour]; p->board_kind == ROOK; p++);
            p->board_kind = ROOK;
        } else if (kind & PROM_BISHOP) {
            for (p = bishop_list[my_colour]; p->board_kind == BISHOP; p++);
            p->board_kind = BISHOP;
        } else {
            for (p = knight_list[my_colour]; p->board_kind == KNIGHT; p++);
            p->board_kind = KNIGHT;
        }
        p->board_where = to;
        (*p->board_count)++;
        board[to] = p;
        UPDATE_KEY(node->node_key, p->board_key[to]);
        node->node_mat[my_colour] += p->board_value;
    } else {
        if (kind == CAPT)
            DO_CAPT(to)
                DO_MOVE(from, to)
                if (kind == CAPT)
                p->board_captured = true;
            else if (kind == DOUBLE)
                hack_ep = to;
                      	
               else if (board[to]->board_kind != PAWN)	
			// RE - added 'if' clause to ensure pawn moves marked non-rev */
			node->node_reverse = true;
	
                
    }

label_null:
    node->node_total_key = node->node_key;
    update_key(your_colour, &(node->node_total_key));
#if DEBUG_LEVEL > 0
    if ((err = check_list()) > 0) {
        my_printf("%s err=%d\n", colour_as_string[my_colour], err);
        move_to_string(m);
        my_printf("%s\n", move_string);
    
        FATAL_ERROR("do list error");
    }
    temp_key = return_cur_key();
    if (!EQUAL_KEY(temp_key, node->node_key)) {
        my_printf("%s\n", colour_as_string[my_colour]);
        move_to_string(m);
        my_printf("%s\n", move_string);
     
        FATAL_ERROR("do key error");
    }
    if (node->node_mat[WHITE_COLOUR] != return_mat_score(WHITE_COLOUR)) {
        my_printf("%s\n", colour_as_string[my_colour]);
        move_to_string(m);
        my_printf("%s\n", move_string);
    
        FATAL_ERROR("do white mat error");
    }
    if (node->node_mat[BLACK_COLOUR] != return_mat_score(BLACK_COLOUR)) {
        my_printf("%s\n", colour_as_string[my_colour]);
        move_to_string(m);
        my_printf("%s\n", move_string);
       FATAL_ERROR("do black mat error");
    }
#endif
    if (my_colour == WHITE_COLOUR)
	{
        BLACK_TO_MOVE;
	}
    else
	{
        WHITE_TO_MOVE;
	}
#undef DO_CAPT
#undef DO_MOVE
}

void            undo_move(void)
{
#define MOVE(A, B)\
        {\
                p = board[A];\
                p->board_where = B;\
                p->board_moved = node->node_moved;\
                p->board_captured = node->node_captured;\
                board[A] = &empty;\
                board[B] = p;\
        }

    board_t        *p;
    long            kind;
    int             from,
                    to;
    int             ep;
#if DEBUG_LEVEL > 0
    int             err;
#endif

    if (my_colour == WHITE_COLOUR)
	{
        BLACK_TO_MOVE;
	}
    else
	{
        WHITE_TO_MOVE;
	}


    /* We have to peek in the previous node for the move. */


    UNPACK_MOVE(node[-1].node_move, kind, from, to)
        if (kind == NULL_MOVE)
        goto label_null;
    if ((kind != SHORT) and(kind != LONG)) {
        BUG(column[from] == INVALID)
            BUG(column[to] == INVALID)
    }
    if (kind == EP) {
        ep = to - pawn_dir[my_colour][0];
        p = board[ep] = node->capt;
        p->board_where = ep;
        (*p->board_count)++;
        MOVE(to, from);
        tot_count[your_colour]++;
    } else if (kind & PROM_CAPT) {
        p = board[to];
        p->board_kind = p->board_where = INVALID;
        (*p->board_count)--;

        p = board[to] = node->capt;
        p->board_where = to;
        tot_count[your_colour]++;
        (*p->board_count)++;

        p = board[from] = node->prom;
        p->board_where = from;
        (*p->board_count)++;
    } else if (kind == SHORT) {
        if (my_colour == WHITE_COLOUR) {
            MOVE(F1, H1);
            MOVE(G1, E1);
        } else {
            MOVE(F8, H8);
            MOVE(G8, E8);
        }
    } else if (kind == LONG) {
        if (my_colour == WHITE_COLOUR) {
            MOVE(D1, A1);
            MOVE(C1, E1);
        } else {
            MOVE(D8, A8);
            MOVE(C8, E8);
        }
    } else if (kind & PROM_NORM) {
        p = board[to];
        p->board_kind = p->board_where = INVALID;
        (*p->board_count)--;
        board[to] = &empty;
        p = board[from] = node->prom;
        p->board_where = from;
        (*p->board_count)++;
    } else {
        MOVE(to, from);
        if (kind == CAPT) {
            p = board[to] = node->capt;
            p->board_where = to;
            tot_count[your_colour]++;
            (*p->board_count)++;
        }
    }
label_null:
    node--;
    inode--;
    BUG(inode < 0)
        hack_ep = node->node_hack_ep;
#if DEBUG_LEVEL > 0
    if ((err = check_list()) > 0) {
        my_printf("%s err=%d\n", colour_as_string[my_colour], err);
        move_to_string(node->node_move);
        my_printf("%s\n", move_string);
    
        FATAL_ERROR("undo list error");
    }
#endif
}

/*
   move_to_san_string converts a move to Short Algebraic Notation
 */

void            move_to_san_string(move_t m, int strict)
{
    int             check;
    long            kind;
    int             from,
                    to;
    char            c;
    int             ikind;
    board_t        *p;
    int             imove,
                    nmoves;
    int             nfound_col,
                    nfound_row;
    int             nlegal;
    move_t         *base;


    /* We check if this is a legal move */



    do_move(m);
    check = i_am_check(your_colour, my_colour);
    undo_move();
    if (check) {
        sprintf(move_string, "??");
        return;
    }
    UNPACK_MOVE(m, kind, from, to)
    /* First some unambigous cases. */


        if (kind & EP) {
        sprintf(move_string, "%cx%s", GET_COL(from), GET_NOTA(to));
    } else if (kind & SHORT) {
        sprintf(move_string, "O-O");
    } else if (kind & LONG) {
        sprintf(move_string, "O-O-O");
    } else if (kind & PROM_CAPT) {
        if (kind & PROM_QUEEN)
            c = 'Q';
        else if (kind & PROM_ROOK)
            c = 'R';
        else if (kind & PROM_BISHOP)
            c = 'B';
        else
            c = 'N';
        sprintf(move_string, "%cx%s=%c", GET_COL(from), GET_NOTA(to), c);
    } else if (kind & PROM_NORM) {
        if (kind & PROM_QUEEN)
            c = 'Q';
        else if (kind & PROM_ROOK)
            c = 'R';
        else if (kind & PROM_BISHOP)
            c = 'B';
        else
            c = 'N';
        sprintf(move_string, "%s=%c", GET_NOTA(to), c);
    } else if (board[from]->board_kind == PAWN) {
        if (kind & CAPT) {
            sprintf(move_string, "%cx%s", GET_COL(from), GET_NOTA(to));
        } else {
            sprintf(move_string, "%s", GET_NOTA(to));
        }
    } else {
        /* Now the ambiguous cases. Ambigous moves are moves by the same kind
         * of piece with the same 'to' field, so we have to generate some
         * moves here. */


        move_top = move_stack;
        base = capt_top;
        ikind = board[from]->board_kind;
        for (p = list[my_colour][ikind]; p->board_kind != INVALID; p++) {
            if (p->board_where == INVALID)
                continue;
            gen_moves(my_colour, your_colour, p);
        }
        nmoves = move_top - move_stack;
        for (imove = 0; imove < nmoves; imove++)
            *capt_top++ = move_stack[imove];
        nmoves = capt_top - base;

        nfound_col = nfound_row = 0;

        for (imove = 0; imove < nmoves; imove++) {
/*
                        Skip the move itself.
                        */

            if (base[imove] == m)
                continue;

            /* Only legal moves */


            do_move(base[imove]);
            check = i_am_check(your_colour, my_colour);
            undo_move();
            if (check)
                continue;

            UNPACK_MOVE(base[imove], other_kind, other_from, other_to);
            if (other_to == to) {
                /* This move seems to be ambiguous. Distinguish by file and
                 * by rank. */
                if (column[other_from] != column[from])
                    nfound_col++;
                else if (row[other_from] != row[from])
                    nfound_row++;
                else
                    FATAL_ERROR("from");
            }
        }
        capt_top = base;

        c = nota_kind[board[from]->board_kind];
        if ((nfound_col == 0) and(nfound_row == 0)) {
            /* Not ambiguous. */


            if (kind & CAPT)
                sprintf(move_string, "%cx%s", c, GET_NOTA(to));
            else
                sprintf(move_string, "%c%s", c, GET_NOTA(to));
        } else if (nfound_col == 1) {
            if (kind & CAPT)
                sprintf(move_string, "%c%cx%s", c, GET_COL(from), GET_NOTA(to));
            else
                sprintf(move_string, "%c%c%s", c, GET_COL(from), GET_NOTA(to));
        } else if (nfound_row == 1) {
            if (kind & CAPT)
                sprintf(move_string, "%c%cx%s", c, GET_ROW(from), GET_NOTA(to));
            else
                sprintf(move_string, "%c%c%s", c, GET_ROW(from), GET_NOTA(to));
        } else {
            if (kind & CAPT)
                sprintf(move_string, "%c%sx%s", c, GET_NOTA(from), GET_NOTA(to));
            else
                sprintf(move_string, "%c%s%s", c, GET_NOTA(from), GET_NOTA(to));
        }
    }
    /* Finally add check or mate */


    if (strict) {
        do_move(m);
        check = i_am_check(my_colour, your_colour);
        if (check) {
            /* Is this a mating move? */

            base = capt_top;
            nlegal = nmoves = gen_all_moves();
            for (imove = 0; imove < nmoves; imove++) {
                do_move(base[imove]);
                check = i_am_check(your_colour, my_colour);
                undo_move();
                if (check)
                    nlegal--;
            }
            capt_top = base;
            if (nlegal == 0)
                strcat(move_string, "#");
            else
                strcat(move_string, "+");
        }
        undo_move();
    }
}

int             search_san(char *move)
{
    int             imove;
    int             nmoves;
    move_t         *base;

    base = capt_top;
    nmoves = gen_all_moves();
    for (imove = 0; imove < nmoves; imove++) {
        move_to_san_string(base[imove], true);
        if (strcmp(move, move_string) == 0)
            break;
    }
    if (imove == nmoves)
        return (INVALID);
    return (imove);
}

/*
   Score simple endgames.
 */

static int       return_endgame_score(int my_colour, int your_colour)
{
    int             drow,
                    dcol,
                    dking;
    int             ikind;
    board_t        *p;

    BUG((tot_count[WHITE_COLOUR] + tot_count[BLACK_COLOUR]) > 3)
        drow = row[king_list[my_colour]->board_where] - row[king_list[your_colour]->board_where];
    dcol = column[king_list[my_colour]->board_where] - column[king_list[your_colour]->board_where];
    dking = ABS(drow) + ABS(dcol);

    if (tot_count[your_colour] == 1) {
        if (tot_count[my_colour] == 1)
            return (0);
        /* Find piece. */
        for (ikind = PAWN; ikind <= QUEEN; ikind++) {
            for (p = list[my_colour][ikind]; p->board_kind != INVALID; p++)
                if (p->board_where != INVALID)
                    goto foundx;
        }
foundx:
        if (p->board_kind == PAWN)
            return (free_pawn_grad[my_colour][p->board_where] - 0);
        else if (p->board_kind == KNIGHT)
            return (0);
        else if (p->board_kind == BISHOP)
            return (0);
        else if (p->board_kind == ROOK)
            return (SCORE_ENDGAME - dking - king_alone_grad[king_list[your_colour]->board_where]);
        else if (p->board_kind == QUEEN)
            return (SCORE_ENDGAME - dking - king_alone_grad[king_list[your_colour]->board_where]);
        else
            FATAL_ERROR("board error");
    } else {
        /* Find piece. */


        for (ikind = PAWN; ikind <= QUEEN; ikind++) {
            for (p = list[your_colour][ikind]; p->board_kind != INVALID; p++)
                if (p->board_where != INVALID)
                    goto foundy;
        }
foundy:
        if (p->board_kind == PAWN)
            return (0 - free_pawn_grad[your_colour][p->board_where]);
        else if (p->board_kind == KNIGHT)
            return (0);
        else if (p->board_kind == BISHOP)
            return (0);
        else if (p->board_kind == ROOK)
            return (dking + king_alone_grad[king_list[my_colour]->board_where] - SCORE_ENDGAME);
        else if (p->board_kind == QUEEN)
            return (dking + king_alone_grad[king_list[my_colour]->board_where] - SCORE_ENDGAME);
        else
            FATAL_ERROR("board error");
    }
    return (SCORE_INVAL);
}

static void      set_attack_board(int my_colour, int your_colour)
{
    board_t        *p;
    int             ikind;
    int            *d,
                    j;
    int             icolumn;

    memcpy(attack_board + A8, attack_zero + A8, (H1 - A8 + 1) * sizeof(int));
    for (icolumn = 0; icolumn < 8; icolumn++)
        occup[WHITE_COLOUR][icolumn] = occup[BLACK_COLOUR][icolumn] = 0;
    for (ikind = PAWN; ikind <= KING; ikind++) {
        for (p = list[my_colour][ikind]; p->board_kind != INVALID; p++) {
            if (p->board_where == INVALID)
                continue;
            if (p->board_kind == PAWN) {
                occup[my_colour][column[p->board_where]] |= BIT(row[p->board_where]);
                j = p->board_where + pawn_dir[my_colour][1];
                attack_board[j]++;
                j = p->board_where + pawn_dir[my_colour][2];
                attack_board[j]++;
            } else if (!p->board_shift) {
                for (d = p->board_dir; *d != 0; d++) {
                    j = p->board_where + *d;
                    attack_board[j]++;
                }
            } else {
                for (d = p->board_dir; *d != 0; d++) {
                    for (j = p->board_where + *d; board[j] == &empty; j += *d)
                        attack_board[j]++;
                    attack_board[j]++;
                }
            }
        }

        /* Same for your colour */


        for (p = list[your_colour][ikind]; p->board_kind != INVALID; p++) {
            if (p->board_where == INVALID)
                continue;
            if (p->board_kind == PAWN) {
                occup[your_colour][column[p->board_where]] |= BIT(row[p->board_where]);
                j = p->board_where + pawn_dir[your_colour][1];
                attack_board[j]--;
                j = p->board_where + pawn_dir[your_colour][2];
                attack_board[j]--;
            } else if (!p->board_shift) {
                for (d = p->board_dir; *d != 0; d++) {
                    j = p->board_where + *d;
                    attack_board[j]--;
                }
            } else {
                for (d = p->board_dir; *d != 0; d++) {
                    for (j = p->board_where + *d; board[j] == &empty; j += *d)
                        attack_board[j]--;
                    attack_board[j]--;
                }
            }
        }
    }
}

/*
   set_score sets the various positional parameters. Note: the assignment truncates.
   This causes no problem when optimizing evaluation function parameters.
   If the best value of for example weak_queen_score would be 10, than
   y[0] would converge to a value between 10.0 and 10.99 if truncation is used,
   but to a value between 9.5 and 10.49 if normal rounding was used.
 */

void            set_score(double *y)
{
   
   
    weak_queen_score = y[0];
    weak_rook_score = y[1];
    weak_bishop_score = y[2];
    weak_knight_score = y[3];
    isolated_pawn_score = y[4];
    king_danger_score = y[5];
    double_pawn_score = y[6];
    lagged_pawn_score = y[7];
    bishop_blocks_pawn_score = y[8];
    bishop_not_developed_score = y[9];
    knight_not_developed_score = y[10];
    weak_pawn_attack_score = y[11];
    pin_score = y[12];
    rook_on_half_open_file_score = y[13];
    knight_strong_score = y[14];
    bishop_strong_score = y[15];
    rook_on_open_file_score = y[16];
    rook_on_7th_rank_score = y[17];
    long_castle_score = y[18];
    short_castle_score = y[19];
    knight_can_move_score = y[20];
    bishop_can_move_score = y[21];
    rook_can_move_score = y[22];
}

/*
   return_pawn_pos_score analyzes the pawn structure. pawn_board is set
   if the pawn is weak.
 */

static int       return_pawn_pos_score(int my_colour, int your_colour, int verbose)
{
    board_t        *p;
    int             icolumn,
                    irow;
    int             sum = 0;
    int             n;
    int             tst;

    for (p = pawn_list[my_colour]; p->board_kind != INVALID; p++) {
        if (p->board_where == INVALID)
            continue;
        sum += pawn_visit[my_colour][p->board_where];
        pawn_board[p->board_where] = false;
        icolumn = column[p->board_where];
        irow = row[p->board_where];
        if (bit_count[occup[my_colour][icolumn]] > 1) {
            if (verbose)
                my_printf("%spawn on %s is doubled\n",
                     colour_as_string[my_colour], GET_NOTA(p->board_where));
            sum += double_pawn_score;
        }
        /* For white: we have a free pawn if the row is greater than or equal
         * to the highest bit of the occupation numbers of the current, left
         * and right row. */


        if (icolumn == 0)
            n = occup[your_colour][1];
        else if (icolumn == 7)
            n = occup[your_colour][6];
        else
            n = (occup[your_colour][icolumn - 1] | occup[your_colour][icolumn + 1]);
        n |= (occup[your_colour][icolumn] | occup[my_colour][icolumn]);
        if (my_colour == WHITE_COLOUR)
            tst = (irow >= bit_high[n]);
        else
            tst = (irow <= bit_low[n]);
        if (tst) {
            if (verbose)
                my_printf("%spawn on %s is free\n",
                     colour_as_string[my_colour], GET_NOTA(p->board_where));
            sum += free_pawn_grad[my_colour][p->board_where];
            continue;
        }
        /* We have an isolated pawn if there are no pawns to the left and to
         * the right. */


        if (icolumn == 0)
            n = occup[my_colour][1];
        else if (icolumn == 7)
            n = occup[my_colour][6];
        else
            n = (occup[my_colour][icolumn - 1] | occup[my_colour][icolumn + 1]);
        if (n == 0) {
            if (verbose)
                my_printf("%spawn on %s is isolated\n",
                     colour_as_string[my_colour], GET_NOTA(p->board_where));
            sum += isolated_pawn_score;
            pawn_board[p->board_where] = true;
            continue;
        }
        /* This is just a poor check for a lagged pawn. */


        if (my_colour == WHITE_COLOUR)
            tst = (irow < (bit_low[n] - 1));
        else
            tst = (irow > (bit_high[n] + 1));
        if (tst) {
            if (verbose)
                my_printf("%spawn on %s is lagged\n",
                     colour_as_string[my_colour], GET_NOTA(p->board_where));
            sum += lagged_pawn_score;
            pawn_board[p->board_where] = true;
        }
    }
    return (sum);
}

/*
   return_pos_score analyzes the remaining pieces. Since the attack board
   is generated only once (for the colour to move), we have to check either >= 1
   or <= -1. This is controlled by the flag inverse).
 */

static int       return_pos_score(int my_colour, int your_colour, int verbose, int inverse)
{
#define MY_PAWN(X) ((board[X]->board_colour == my_colour) and (board[X]->board_kind == PAWN))
#define MY_ROOK(X) ((board[X]->board_colour == my_colour) and (board[X]->board_kind == ROOK))
    board_t        *p,
                   *q;
    int             tst;
    int             icolumn,
                    irow;
    int             n,
                    m;
    int             sum = 0;
    int             j,
                    k;
    int            *d;

    for (p = knight_list[my_colour]; p->board_kind != INVALID; p++) {
        if (p->board_where == INVALID)
            continue;
        sum += knight_visit[my_colour][p->board_where];
        if (!p->board_moved)
            sum += knight_not_developed_score;
        irow = row[p->board_where];
        icolumn = column[p->board_where];


        /* A knight is strong if it is a 'free pawn' and covered by one or
         * (even better) two pawns */

        if (icolumn == 0)
            n = occup[your_colour][1];
        else if (icolumn == 7)
            n = occup[your_colour][6];
        else
            n = (occup[your_colour][icolumn - 1] | occup[your_colour][icolumn + 1]);
        if (my_colour == WHITE_COLOUR)
            tst = ((n == 0) or(irow >= bit_high[n]));
        else
            tst = ((n == 0) or(irow <= bit_low[n]));
        if (tst) {
            q = board[p->board_where - pawn_dir[my_colour][1]];
            if ((q->board_colour == my_colour) and(q->board_kind == PAWN)) {
                if (verbose)
                    my_printf("%sknight on %s is strong\n",
                     colour_as_string[my_colour], GET_NOTA(p->board_where));
                sum += knight_strong_score;
            }
            q = board[p->board_where - pawn_dir[my_colour][2]];
            if ((q->board_colour == my_colour) and(q->board_kind == PAWN)) {
                if (verbose)
                    my_printf("%sknight on %s is strong\n",
                     colour_as_string[my_colour], GET_NOTA(p->board_where));
                sum += knight_strong_score;
            }
        }
        /* We define mobility as the number of squares to which the knight
         * can move without the risk of being captured. n is the number of
         * squares controlled by my_colour, m is the number of squares
         * controlled by your_colour. Note: we have to check if attack_board
         * >= 1, not >= 0 because the piece itself is counted when generating
         * the attack board. */


        n = m = 0;
        for (d = p->board_dir; *d != 0; d++) {
            j = p->board_where + *d;
            q = board[j];
            if (q == &empty) {
                if (inverse)
                    tst = (attack_board[j] <= -1);
                else
                    tst = (attack_board[j] >= 1);
                if (tst)
                    n += knight_can_move_score;
                else
                    m += knight_can_move_score;
            } else if (q->board_colour == your_colour) {
                if (q->board_kind == PAWN) {
                    if (pawn_board[j]) {
                        if (verbose)
                            my_printf("%sknight on %s attacks weak pawn on %s\n",
                                      colour_as_string[my_colour], GET_NOTA(p->board_where), GET_NOTA(j));
                        sum += weak_pawn_attack_score;
                    }
                }
            }
        }
        /* A knight is weak if it can only move to sqares controlled by the
         * opponent. */

        if ((n == 0) and(m > 0)) {
            if (verbose)
                my_printf("%sknight on %s is weak\n",
                     colour_as_string[my_colour], GET_NOTA(p->board_where));
            sum += weak_knight_score;
        } else
            sum += (n - m);
    }


    /* Same for bishop and rook. */

    for (p = bishop_list[my_colour]; p->board_kind != INVALID; p++) {
        if (p->board_where == INVALID)
            continue;
        if (!p->board_moved)
            sum += bishop_not_developed_score;
        irow = row[p->board_where];
        icolumn = column[p->board_where];
        j = p->board_where - pawn_dir[my_colour][0];
        q = board[j];
        if ((q->board_colour == my_colour) and(q->board_kind == PAWN)) {
            if (verbose)
                my_printf("%sbishop on %s blocks pawn on %s\n",
                          colour_as_string[my_colour], GET_NOTA(p->board_where), GET_NOTA(j));
            sum += bishop_blocks_pawn_score;
        }
        if (icolumn == 0)
            n = occup[your_colour][1];
        else if (icolumn == 7)
            n = occup[your_colour][6];
        else
            n = (occup[your_colour][icolumn - 1] | occup[your_colour][icolumn + 1]);
        if (my_colour == WHITE_COLOUR)
            tst = ((n == 0) or(irow >= bit_high[n]));
        else
            tst = ((n == 0) or(irow <= bit_low[n]));
        if (tst) {
            q = board[p->board_where - pawn_dir[my_colour][1]];
            if ((q->board_colour == my_colour) and(q->board_kind == PAWN)) {
                if (verbose)
                    my_printf("%sbishop on %s is strong\n",
                     colour_as_string[my_colour], GET_NOTA(p->board_where));
                sum += bishop_strong_score;
            }
            q = board[p->board_where - pawn_dir[my_colour][2]];
            if ((q->board_colour == my_colour) and(q->board_kind == PAWN)) {
                if (verbose)
                    my_printf("%sbishop on %s is strong\n",
                     colour_as_string[my_colour], GET_NOTA(p->board_where));
                sum += bishop_strong_score;
            }
        }
        n = m = 0;
        for (d = p->board_dir; *d != 0; d++) {
            for (j = p->board_where + *d; board[j] == &empty; j += *d) {
                if (inverse)
                    tst = (attack_board[j] <= -1);
                else
                    tst = (attack_board[j] >= 1);
                if (tst)
                    n += bishop_can_move_score;
                else
                    m += bishop_can_move_score;
            }
            q = board[j];
            if (q->board_colour == your_colour) {
                if (q->board_kind == PAWN) {
                    if (pawn_board[j]) {
                        if (verbose)
                            my_printf("%sbishop on %s attacks weak pawn on %s\n",
                                      colour_as_string[my_colour], GET_NOTA(p->board_where), GET_NOTA(j));
                        sum += weak_pawn_attack_score;
                    }
                } else {
                    for (k = q->board_where + *d; board[k] == &empty; k += *d);
                    q = board[k];
                    if ((q->board_colour == your_colour) and(q->board_kind == KING)) {
                        if (verbose)
                            my_printf("%sbishop on %s pins piece on %s\n",
                                      colour_as_string[my_colour], GET_NOTA(p->board_where), GET_NOTA(j));
                        sum += pin_score;
                    }
                }
            }
        }
        if ((n == 0) and(m > 0)) {
            if (verbose)
                my_printf("%sbishop on %s is weak\n",
                     colour_as_string[my_colour], GET_NOTA(p->board_where));
            sum += weak_bishop_score;
        } else
            sum += (n - m);
    }
    for (p = rook_list[my_colour]; p->board_kind != INVALID; p++) {
        if (p->board_where == INVALID)
            continue;
        irow = row[p->board_where];
        icolumn = column[p->board_where];
        n = (occup[my_colour][icolumn] | occup[your_colour][icolumn]);
        if (n == 0) {
            if (verbose)
                my_printf("%srook on %s is on open file\n",
                     colour_as_string[my_colour], GET_NOTA(p->board_where));
            sum += rook_on_open_file_score;
        } else if ((occup[my_colour][icolumn] == 0) or
                   (occup[your_colour][icolumn] == 0)) {
            if (verbose)
                my_printf("%srook on %s is on half open file\n",
                     colour_as_string[my_colour], GET_NOTA(p->board_where));
            sum += rook_on_half_open_file_score;
        }
        if (irow == strong_row[my_colour]) {
            if (verbose)
                my_printf("%srook on %s is on 7th rank\n",
                     colour_as_string[my_colour], GET_NOTA(p->board_where));
            sum += rook_on_7th_rank_score;
        }
        n = m = 0;
        for (d = p->board_dir; *d != 0; d++) {
            for (j = p->board_where + *d; board[j] == &empty; j += *d) {
                if (inverse)
                    tst = (attack_board[j] <= -1);
                else
                    tst = (attack_board[j] >= 1);
                if (tst)
                    n += rook_can_move_score;
                else
                    m += rook_can_move_score;
            }
            q = board[j];
            if (q->board_colour == your_colour) {
                if (q->board_kind == PAWN) {
                    if (pawn_board[j]) {
                        if (verbose)
                            my_printf("%srook on %s attacks weak pawn on %s\n",
                                      colour_as_string[my_colour], GET_NOTA(p->board_where), GET_NOTA(j));
                        sum += weak_pawn_attack_score;
                    }
                } else {
                    for (k = q->board_where + *d; board[k] == &empty; k += *d);
                    q = board[k];
                    if ((q->board_colour == your_colour) and(q->board_kind == KING)) {
                        if (verbose)
                            my_printf("%srook on %s pins piece on %s\n",
                                      colour_as_string[my_colour], GET_NOTA(p->board_where), GET_NOTA(j));
                        sum += pin_score;
                    }
                }
            }
        }
        if ((n == 0) and(m > 0)) {
            if (verbose)
                my_printf("%sweak rook on %s\n",
                     colour_as_string[my_colour], GET_NOTA(p->board_where));
            sum += weak_rook_score;
        } else
            sum += (n - m);
    }
    for (p = queen_list[my_colour]; p->board_kind != INVALID; p++) {
        if (p->board_where == INVALID)
            continue;
        n = m = 0;
        for (d = p->board_dir; *d != 0; d++) {
            for (j = p->board_where + *d; board[j] == &empty; j += *d) {
                if (inverse)
                    tst = (attack_board[j] <= -1);
                else
                    tst = (attack_board[j] >= 1);
                if (tst)
                    n++;
                else
                    m++;
            }
            q = board[j];
            if (q->board_colour == your_colour) {
                if (q->board_kind == PAWN) {
                    if (pawn_board[j]) {
                        if (verbose)
                            my_printf("%squeen on %s attacks weak pawn on %s\n",
                                      colour_as_string[my_colour], GET_NOTA(p->board_where), GET_NOTA(j));
                        sum += weak_pawn_attack_score;
                    }
                } else {
                    for (k = q->board_where + *d; board[k] == &empty; k += *d);
                    q = board[k];
                    if ((q->board_colour == your_colour) and(q->board_kind == KING)) {
                        if (verbose)
                            my_printf("%squeen on %s pins piece on %s\n",
                                      colour_as_string[my_colour], GET_NOTA(p->board_where), GET_NOTA(j));
                        sum += pin_score;
                    }
                }
            }
        }
        if ((n == 0) and(m > 0)) {
            if (verbose)
                my_printf("%sweak queen on %s\n",
                     colour_as_string[my_colour], GET_NOTA(p->board_where));
            sum += weak_queen_score;
        }
    }

    /* King safety */


    p = king_list[my_colour];
    for (d = p->board_dir; *d != 0; d++) {
        j = p->board_where + *d;
        if (board[j] == &side)
            continue;
        if (inverse)
            tst = (attack_board[j] > 0);
        else
            tst = (attack_board[j] < 0);
        if (tst) {
            if (verbose)
                my_printf("%sking on %s is in danger on %s\n",
                          colour_as_string[my_colour], GET_NOTA(p->board_where), GET_NOTA(j));
            sum += king_danger_score;
        }
    }

    /* We assume castling is important if the opponent has at least a queen
     * and a rook */


    if ((piece_count[your_colour][QUEEN] > 0) and(piece_count[your_colour][ROOK] > 0)) {
        /* Very poor castle check */


        if (my_colour == WHITE_COLOUR) {
            if ((p->board_where == G1) or(p->board_where == H1) or
                (p->board_where == G2) or(p->board_where == H2)) {
                n = m = 0;
                if (MY_PAWN(F2))
                    n++;
                if (MY_PAWN(G2))
                    n++;
                if (MY_PAWN(H2))
                    n++;
                if (MY_PAWN(F3))
                    n++;
                if (MY_PAWN(G3))
                    n++;
                if (MY_PAWN(H3))
                    n++;
                if (MY_ROOK(G1))
                    m++;
                if (MY_ROOK(H1))
                    m++;
                if (MY_ROOK(G2))
                    m++;
                if (MY_ROOK(H2))
                    m++;
                if ((m == 0) and(n > 2)) {
                    if (verbose)
                        my_printf("%sking on %s has short castled\n",
                                  colour_as_string[my_colour], GET_NOTA(p->board_where));
                    sum += short_castle_score;
                }
            } else if ((p->board_where == C1) or
                       (p->board_where == B1) or(p->board_where == A1) or
                       (p->board_where == B2) or(p->board_where == A2)) {
                n = m = 0;
                if (MY_PAWN(C2))
                    n++;
                if (MY_PAWN(B2))
                    n++;
                if (MY_PAWN(A2))
                    n++;
                if (MY_PAWN(C3))
                    n++;
                if (MY_PAWN(B3))
                    n++;
                if (MY_PAWN(A3))
                    n++;
                if (MY_ROOK(B1))
                    m++;
                if (MY_ROOK(A1))
                    m++;
                if (MY_ROOK(B2))
                    m++;
                if (MY_ROOK(A2))
                    m++;
                if ((m == 0) and(n > 2)) {
                    if (verbose)
                        my_printf("%sking on %s has long castled\n",
                                  colour_as_string[my_colour], GET_NOTA(p->board_where));
                    sum += long_castle_score;
                }
            }
        } else {
            if ((p->board_where == G8) or(p->board_where == H8) or
                (p->board_where == G7) or(p->board_where == H7)) {
                n = m = 0;
                if (MY_PAWN(F7))
                    n++;
                if (MY_PAWN(G7))
                    n++;
                if (MY_PAWN(H7))
                    n++;
                if (MY_PAWN(F6))
                    n++;
                if (MY_PAWN(G6))
                    n++;
                if (MY_PAWN(H6))
                    n++;
                if (MY_ROOK(G8))
                    m++;
                if (MY_ROOK(H8))
                    m++;
                if (MY_ROOK(G7))
                    m++;
                if (MY_ROOK(H7))
                    m++;
                if ((m == 0) and(n > 2)) {
                    if (verbose)
                        my_printf("%sking on %s has short castled\n",
                                  colour_as_string[my_colour], GET_NOTA(p->board_where));
                    sum += short_castle_score;
                }
            } else if ((p->board_where == C8) or
                       (p->board_where == B8) or(p->board_where == A8) or
                       (p->board_where == B7) or(p->board_where == A7)) {
                n = m = 0;
                if (MY_PAWN(C7))
                    n++;
                if (MY_PAWN(B7))
                    n++;
                if (MY_PAWN(A7))
                    n++;
                if (MY_PAWN(C6))
                    n++;
                if (MY_PAWN(B6))
                    n++;
                if (MY_PAWN(A6))
                    n++;
                if (MY_ROOK(B8))
                    m++;
                if (MY_ROOK(A8))
                    m++;
                if (MY_ROOK(B7))
                    m++;
                if (MY_ROOK(A7))
                    m++;
                if ((m == 0) and(n > 2)) {
                    if (verbose)
                        my_printf("%sking on %s has long castled\n",
                                  colour_as_string[my_colour], GET_NOTA(p->board_where));
                    sum += long_castle_score;
                }
            }
        }
    }
    
    
    
    // some bonuses based on material differences - JA
    
    if (piece_count[my_colour][QUEEN] == 1  && piece_count[your_colour][QUEEN] == 0) sum += QUEEN_SCORE_BONUS; 
    // I've got a queen advantage - so give a big queen bonus - JA
    
    if (piece_count[my_colour][QUEEN] == 0  && piece_count[your_colour][QUEEN] == 0) {
    if (piece_count[my_colour][ROOK] > 0)   sum += ROOK_SCORE_BONUS;
    if (piece_count[my_colour][KNIGHT] > 0) sum += KNIGHT_SCORE_BONUS;
    if (piece_count[my_colour][BISHOP] > 0) sum += BISHOP_SCORE_BONUS;
    // no queens left on either side so make my other pieces a little more valuable - JA
    }
    
    if (piece_count[my_colour][BISHOP] == 2) sum += BISHOP_SCORE_BONUS;      // for bishop pair - JA
    
    
    if (piece_count[my_colour][QUEEN] == 0  && piece_count[my_colour][ROOK] == 0 
    && piece_count[my_colour][KNIGHT] == 0 &&  piece_count[my_colour][BISHOP] == 0
    && piece_count[my_colour][PAWN] > 0) {
    
    sum += PAWN_SCORE_BONUS;    // I've only only got pawns left so make them more valuable - JA
}                                
    
    
    if (piece_count[my_colour][KNIGHT]  >  piece_count[your_colour][KNIGHT]) sum += KNIGHT_SCORE_BONUS;
    // I've got more knights than you, so give a knight bonus score - JA
    if (piece_count[my_colour][ROOK]  >  piece_count[your_colour][ROOK]) sum += ROOK_SCORE_BONUS; 
    // I've got more rooks than you, so give a rook bonus score - JA
    if (piece_count[my_colour][BISHOP]  >  piece_count[your_colour][BISHOP]) sum +=BISHOP_SCORE_BONUS; 
    // I've got more bishops than you, so give a bishop bonus score - JA
    if (piece_count[my_colour][PAWN]  >  piece_count[your_colour][PAWN]) sum += PAWN_SCORE_BONUS; 
    // I've got more pawns than you , so give a little pawn bonus score - JA
    
                    
    return (sum);

#undef MY_PAWN
#undef MY_ROOK
}



/*
   We check for draw by repetition by checking if the current key has occurred
   before. In the alpha beta loop, we return true when we find a match.
   In the main loop, we must check whether the position has occurred three times.
 */

int             draw_by_repetition(int strict)
{
    node_t         *temp_node;
    int             ncount;

    temp_node = node;
    ncount = 1;
    while (true) {
        /* your move and my move must be reversible. */


        if (!temp_node->node_reverse)
            break;
        temp_node--;
        if (!temp_node->node_reverse)
            break;
        temp_node--;
        if (EQUAL_KEY(node->node_total_key, temp_node->node_total_key)) {
            ncount++;
            if (!strict)  {
                     
                return (true);
                }
                                                     
            else if (ncount > 2) 
                return (true);
                
        }
    }
    return (false);
}




/*
   A difficult problem in any chess program is the quiescense search (the capture
   evaluation). See for example
   G. Schrufer, ICCA vol. 12, no. 1 (1989) pp. 3-9.
   I do it in the following way: any piece that has captured is marked.
   In the quiescense search, I allow QUIES_DEPTH free captures (so any piece
   can capture any other piece), but after the QUIES_DEPTH free captures, only
   captures to pieces which have their capture mark set are allowed. In this way the
   quiescence search will always terminate. Of course higher QUIES_DEPTH values
   give better results, but also makes the program slower, so there is a trade-off
   here.
 */

static           score_t
                quies(int swap, score_t alpha, score_t beta, int depth, var_t * var_base)
{
    score_t         temp_score,
                    best_score;

    score_cache_t  *p;

    int             nmoves;
    int             i;
    int             ikind;
    board_t        *q;
    var_t          *var,
                   *var_top;
    move_t         *base;

#ifdef BUG_DEPTH
    if (inode > inode_max) {
        int             itemp;
        my_printf("inode max=%d\n", inode - inode_root);
        for (itemp = inode_root; itemp < inode; itemp++) {
            move_to_string(node_stack[itemp].node_move);
            my_printf(" %s", move_string);
        }
        my_printf("\n");
        inode_max = inode;
    }
#endif

    var_top = var_base;
    *var_top++ = INVAL_MOVE;

    if (swap) {
        temp_score = alpha;
        alpha = -beta;
        beta = -temp_score;
    }
    base = capt_top;

    best_score = node->node_mat[my_colour] - node->node_mat[your_colour];
    if (UsePos == 0)
        goto label_skip;

    if ((tot_count[WHITE_COLOUR] + tot_count[BLACK_COLOUR]) < 4) {
        best_score += return_endgame_score(my_colour, your_colour);
        goto label_skip;
    }
    p = score_cache[my_colour] + (node->node_total_key.low & SCORE_CACHE_MASK);
    if ((p->cache_score != SCORE_INVAL) and
        (EQUAL_KEY(p->cache_key, node->node_total_key))) {
#ifdef CACHE_FEN
        char            fen_string[FEN_STRING_MAX];
        board_to_fen(fen_string, false);
        BUG(strcmp(fen_string, p->cache_fen) != 0)
#endif
            best_score += p->cache_score;
        nhits++;
    } else {
        set_attack_board(my_colour, your_colour);
        temp_score = 0;
        temp_score += return_pawn_pos_score(my_colour, your_colour, false);
        temp_score -= return_pawn_pos_score(your_colour, my_colour, false);
        temp_score += return_pos_score(my_colour, your_colour, false, false);
        temp_score -= return_pos_score(your_colour, my_colour, false, true);

        best_score += temp_score;
        nevals++;
        p->cache_score = temp_score;
        p->cache_key = node->node_total_key;
#ifdef CACHE_FEN
        board_to_fen(p->cache_fen, false);
#endif
    }


    /* If by doing nothing we already have a fail high, we do not have to
     * check the captures. */


label_skip:
    if (best_score >= beta)
        goto label_neg;

    /* The captures. We generate the captures to your most valuable pieces
     * first. */



    for (ikind = QUEEN; ikind >= PAWN; ikind--) {
        for (q = list[your_colour][ikind]; q->board_kind != INVALID; q++) {
            if (q->board_where == INVALID)
                continue;


            /* Only recaptures if (depth >= QUIES_DEPTH) */


           if ((depth >= QUIES_DEPTH) and ! q->board_captured)
         
                continue;
            capt_top = base;
            nmoves = gen_capts(q->board_where, my_colour);
            for (i = 0; i < nmoves; i++) {
                do_move(base[i]);
                if (i_am_check(your_colour, my_colour)) {
                    undo_move();
                    continue;
                }
                temp_score = quies(true, MAX(alpha, best_score), beta, depth + 1, var_top);
                undo_move();
                if (temp_score > best_score) {
                    best_score = temp_score;
                    var = var_base;
                    *var++ = base[i];
                    while ((*var++ = *var_top++) != INVAL_MOVE);
                    var_top = var;
                    var = var_base;
                    if (best_score >= beta)
                        goto label_neg;
                }
            }
        }
    }

label_neg:
    capt_top = base;
    if (swap)
        best_score = -best_score;
    return (best_score);
}

/*
   A standard alpha beta routine, with the null move heuristic. See for example
   T.A. Marsland, ICCA vol. 9, no. 1 (1986) pp. 3-19.
   For the null move heuristic see
   Chr. Donninger, ICCA vol. 12, no 3 (1993) pp. 137-143.
   Null moves are used as a forward pruning mechanism: if the null move gives
   a beta cutoff, it is assumed that the regular moves in that position, when searched,
   would also give a beta cutoff. This speeds up the program by a factor of two, but
   it fails terribly in case of zugzwang, because in that case not doing a move is a
   usually a good idea: the null move gives a beta cutoff, but the regular moves will not.
   I call this 'null move failures'.
   It is very easy to check how often null move failures occur: instead of
   doing the null move cutoff, you set a flag. If you do not get a beta cutoff
   when searching the regular moves and that flag is set, you know you have
   a null move failure. I tried this for the 300 win-at-chess positions
   from Lang&Smith, searched with DepthMax = 5. In 5 out of the 300 positions,
   I got null move failure warnings. It still found the right solution
   in 4 of the 5 positions, and in 1 position it did not find the solution any more.
 */



#if DEBUG_LEVEL > 0
static           score_t
                debug_alpha_beta(int null_move, score_t alpha, score_t beta, int depth, var_t * var_base)
{
#define RETURN(X)\
        {best_score = (X); goto label_neg;}

    score_t         best_score,
                    temp_score;
    int             best_move;
    int             check;
    int             nmoves,
                    nlegal;
    int             imove,
                    jmove;
    int             depth_next;
    var_t          *var,
                   *var_top;
    move_t         *base;
    
        
    node_t         *node_debug = node;
    
    
    BUG(alpha >= beta);
    BUG(depth < 0);
    var_top = var_base;
    *var_top++ = INVAL_MOVE;
    
    
    clock_end = clock();
    if (clock_end < clock_begin) {
        my_printf("clock time cycle detected!\n");
        clock_begin = clock_end;
    } else {
        if ((clock_end - clock_begin) > (CLOCKS_PER_SEC / 10)) {
            time_used += (double) (clock_end - clock_begin) / CLOCKS_PER_SEC;
            clock_begin = clock_end;
            if (time_used > stop_time)
                return (SCORE_INVAL);
        }
    }

    temp_score = alpha;
    alpha = -beta;
    beta = -temp_score;

    base = capt_top;

    best_score = -SCORE_LIMIT;
    best_move = INVALID;

    if (draw_by_repetition(false))
        RETURN(0);
    check = i_am_check(my_colour, your_colour);

    if (!check and(depth == 0)) {
#ifdef BUG_DEPTH
        bug_depth[inode - inode_root]++;
#endif
        best_score = quies(false, alpha, beta, 0, var_base);
        goto label_neg;
    }
    if (!check and ! null_move)
        (depth >= NULL_MOVE_DEPTH) and(node->node_mat[my_colour] >= (2 * ROOK_SCORE))) {
        do_move(NULL_MOVE);
        depth_next = depth - 1;
        temp_score = debug_alpha_beta(true, beta - 1, beta, depth_next - NULL_MOVE_REDUC, var_top);
        undo_move();
        if (temp_score == SCORE_INVAL)
            return (SCORE_INVAL);
        if (temp_score >= beta)
            RETURN(temp_score);
    }
    nlegal = 0;
    if (check)
        depth_next = depth;
    else
        depth_next = depth - 1;

    ngens++;
    nmoves = gen_all_moves();

    if (best_move == INVALID)
        best_move = 0;
    else if (best_move >= nmoves) {
        my_printf("hash fault!\n");
        best_move = 0;
    }
    imove = INVALID;
    while (imove < nmoves) {
        if (imove == INVALID)
            jmove = best_move;
        else
            jmove = imove;
        if (base[jmove] == INVAL_MOVE) {
            imove++;
            continue;
        }
        do_move(base[jmove]);
        if (i_am_check(your_colour, my_colour)) {
            undo_move();
            base[jmove] = INVAL_MOVE;
            imove++;
            continue;
        }
        nlegal++;
        ndos++;
        temp_score = debug_alpha_beta(false, MAX(alpha, best_score), beta, depth_next, var_top);
        undo_move();
        if (temp_score == SCORE_INVAL)
            return (SCORE_INVAL);
        if (temp_score > best_score) {
            best_score = temp_score;
            best_move = jmove;
            var = var_base;
            *var++ = base[jmove];
            while ((*var++ = *var_top++) != INVAL_MOVE);
            var_top = var;
            var = var_base;
            if (best_score >= beta)
                goto label_neg;
        }
        base[jmove] = INVAL_MOVE;
        imove++;
    }

    if (nlegal == 0) {
        if (check)
            RETURN(-SCORE_MATE + inode);
        else
            RETURN(0);
    }
label_neg:
    BUG(best_score == -SCORE_LIMIT);
    BUG(node != node_debug);
    capt_top = base;
    return (-best_score);
#undef RETURN
}
#endif

static score_t   alpha_beta(int null_move, score_t alpha, score_t beta, int depth, var_t * var_base)
{
#define RETURN(X)\
        {best_score = (X); goto label_neg;}

    score_t         best_score,
                    temp_score;

    cache_t        *p;

    int             best_move;
    int             check;
    int             nmoves,
                    nlegal;
    int             imove,
                    jmove;
    int             depth_next;
    var_t          *var,
                   *var_top;
    move_t         *base;
   
#if DEBUG_LEVEL > 0
    node_t         *node_debug = node;
    int             debug_score;
#endif


    BUG(alpha >= beta);
    BUG(depth < 0);
    var_top = var_base;
    *var_top++ = INVAL_MOVE;

    /* A previous version of my program occasionally stuck in infinite loops
     * when run on an RS6000. After some debugging, I found out that it had
     * to do with the time control: The code looked something like: t1 =
     * clock(); .. t2 = clock(); if ((t2 - t1) > (TimeLimit *
     * CLOCKS_PER_SEC)) return..
     * 
     * Looks good, but only works on machines with a slow clock (like my MAC).
     * On an RS6000 CLOCKS_PER_SEC is 1000000L, and clock_t is a long, so the
     * clock cycles in about 40 minutes. Now suppose t1 is LONG_MAX (or near
     * LONG_MAX), then t2 is always less than t1 and the program will not
     * terminate! To overcome this problem, I now update the time used every
     * 1/10 of a second. When a clock cycle is detected a warning is printed
     * (no attempt is made to recover from the clock cycle, so the time used
     * may by off by at most 1/10 of a second). */



 
    clock_end = clock();
    if (clock_end < clock_begin) {
        my_printf("clock time cycle detected!\n");
        clock_begin = clock_end;
    } else {
        if ((clock_end - clock_begin) > (CLOCKS_PER_SEC / 10)) {
            time_used += (double) (clock_end - clock_begin) / CLOCKS_PER_SEC;
            clock_begin = clock_end;

            if (time_used > stop_time)
                return (SCORE_INVAL);

        }
    }

    /* I swap alpha and beta here to keep the call to alpha_beta readable. */


    temp_score = alpha;
    alpha = -beta;
    beta = -temp_score;

    base = capt_top;

    best_score = -SCORE_LIMIT;
    best_move = INVALID;


    /* We first check the hash table. Since we use the lower CACHE_KIND_BITS
     * bits to store some information, we do not check these bits. */


    p = cache[my_colour] + (node->node_total_key.low & CACHE_MASK);
    if (!(p->cache_key.low & CACHE_INVAL) and
        (p->cache_key.high == node->node_total_key.high) and
#ifdef HASH_KEY128
        (p->cache_key.lower == node->node_total_key.lower) and
        (p->cache_key.higher == node->node_total_key.higher) and
#endif
        ((p->cache_key.low & ~CACHE_KIND_MASK) == (node->node_total_key.low & ~CACHE_KIND_MASK))) {
#ifdef CACHE_FEN
        char            fen_string[FEN_STRING_MAX];
        board_to_fen(fen_string, false);
        BUG(strcmp(fen_string, p->cache_fen) != 0);
#endif
        best_move = p->cache_best;


        /* Checking if cache_depth >= depth is possibly dangerous. We
         * certainly need to check cache_depth == depth if DEBUG_LEVEL > 1,
         * otherwise the comparison with debug_alpha_beta fails. */

#if DEBUG_LEVEL > 1
        if ((p->cache_key.low & CACHE_DEPTH_MASK) == depth)
#else
        if ((p->cache_key.low & CACHE_DEPTH_MASK) >= depth)
#endif
        {
            if (p->cache_key.low & CACHE_true)
			{
                RETURN(p->cache_score);
			}
            else if (p->cache_key.low & CACHE_LOWER) {
                if (p->cache_score >= beta)
				{
                    RETURN(p->cache_score);
				}
                else if (p->cache_score > alpha)
                    alpha = p->cache_score;
            } else if (p->cache_key.low & CACHE_UPPER) {
                if (p->cache_score <= alpha)
				{
                    RETURN(p->cache_score);
				}
                else if (p->cache_score < beta)
                    beta = p->cache_score;
            }
        }
    }
    if (draw_by_repetition(false))
        RETURN(0);
    check = i_am_check(my_colour, your_colour);


    /* We have an end node if we are not check and depth == 0. */

    if (!check and(depth == 0)) {
#ifdef BUG_DEPTH
        bug_depth[inode - inode_root]++;
#endif
        best_score = quies(false, alpha, beta, 0, var_base);
        goto label_neg;
    }
    /* Try the null move. We assume no zugzwang if material >=
     * (2*ROOK_SCORE). */


    if (!check and ! null_move and
        (depth >= NULL_MOVE_DEPTH) and(node->node_mat[my_colour] >= (2 * ROOK_SCORE))) {
        do_move(NULL_MOVE);
        depth_next = depth - 1;
        temp_score = alpha_beta(true, beta - 1, beta, depth_next - NULL_MOVE_REDUC, var_top);
        undo_move();
        if (temp_score == SCORE_INVAL)
            return (SCORE_INVAL);
        if (temp_score >= beta)
            RETURN(temp_score);
    }
    /* Since we generate pseudo legal moves, we have to keep track of the
     * number of legal moves. */


    nlegal = 0;
    if (check)
        depth_next = depth;
    else
        depth_next = depth - 1;

    /* Generate all moves. */

    ngens++;
    nmoves = gen_all_moves();

    /* Search the moves, the move from the hash table first. */


    if (best_move == INVALID)
        best_move = 0;
    else if (best_move >= nmoves) {
        my_printf("hash fault!\n");
        best_move = 0;
    }

    imove = INVALID;
    while (imove < nmoves) {


        if (imove == INVALID)
            jmove = best_move;
        else
            jmove = imove;
        if (base[jmove] == INVAL_MOVE) {
            imove++;
            continue;
        }
        do_move(base[jmove]);
        if (i_am_check(your_colour, my_colour)) {
            undo_move();
            base[jmove] = INVAL_MOVE;
            imove++;
            continue;
        }
        nlegal++;
        ndos++;
        temp_score = alpha_beta(false, MAX(alpha, best_score), beta, depth_next, var_top);
        undo_move();

        if (temp_score == SCORE_INVAL)
            return (SCORE_INVAL);

        if (temp_score > best_score) {

            best_score = temp_score;
            best_move = jmove;
            var = var_base;
            *var++ = base[jmove];
            while ((*var++ = *var_top++) != INVAL_MOVE);
            var_top = var;
            var = var_base;
            if (best_score >= beta)
                goto label_store;
        }
        base[jmove] = INVAL_MOVE;
        imove++;

    }


    /* When detecting check mate, we add inode to the mate score to make sure
     * we find the shortest mate. */


    if (nlegal == 0) {
        if (check)
		{
            RETURN(-SCORE_MATE)+ inode;  // JA
        
		}
        else
		{
            RETURN(0);
		}
    }
label_store:

#if DEBUG_LEVEL > 1


    /* Sometimes inverting scores is nice, sometimes it just isn't.. */


    debug_score = debug_alpha_beta(false, -beta, -alpha, depth, var_top);
    BUG(debug_score == SCORE_INVAL);
    debug_score = -debug_score;
    if (best_score >= beta) {
        if (debug_score < beta) {
            my_printf("alpha=%d beta=%d best_score=%d debug_score=%d\n",
                      alpha, beta, best_score, debug_score);
            FATAL_ERROR("debug_score < beta");
        }
    } else if (best_score > alpha) {
        if (debug_score >= beta) {
            my_printf("alpha=%d beta=%d best_score=%d debug_score=%d\n",
                      alpha, beta, best_score, debug_score);
            FATAL_ERROR("debug_score >= beta");
        } else if (debug_score < alpha) {
            my_printf("alpha=%d beta=%d best_score=%d debug_score=%d\n",
                      alpha, beta, best_score, debug_score);
            FATAL_ERROR("debug_score < alpha");
        }
    } else {
        if (debug_score > alpha) {
            my_printf("alpha=%d beta=%d best_score=%d debug_score=%d\n",
                      alpha, beta, best_score, debug_score);
            FATAL_ERROR("debug_score > alpha");
        }
    }
#endif



    if (depth > CACHE_DEPTH_MASK)
        goto label_neg;
    p->cache_key.high = node->node_total_key.high;


    /* We stuff the depth and the flags in the lower CACHE_KIND_BITS. */


    p->cache_key.low = ((node->node_total_key.low & ~CACHE_KIND_MASK) | depth);
    p->cache_score = best_score;
    if (best_score <= alpha)
        p->cache_key.low |= CACHE_UPPER;
    else if (best_score >= beta)
        p->cache_key.low |= CACHE_LOWER;
    else
        p->cache_key.low |= CACHE_true;
    p->cache_best = best_move;
#ifdef CACHE_FEN
    board_to_fen(p->cache_fen, false);
#endif

label_neg:
    BUG(best_score == -SCORE_LIMIT);
#if DEBUG_LEVEL > 0
    BUG(node != node_debug);
#endif

    capt_top = base;
    return (-best_score);
#undef RETURN
}


/*
   mate_alpha_beta is a special alpha beta routine for mate finding
 */

static score_t   mate_alpha_beta(score_t alpha, score_t beta, int depth, var_t * var_base)
{
#define RETURN(X)\
        {best_score = (X); goto label_neg;}

    score_t         best_score,
                    temp_score;

    cache_t        *p;

    int             best_move;
    int             check;
    int             nmoves,
                    nlegal;
    int             imove,
                    jmove;
    var_t          *var,
                   *var_top;
    move_t         *base;
    


    BUG(alpha >= beta);
    BUG(depth < 0);
    var_top = var_base;
    *var_top++ = INVAL_MOVE;
    
   
    clock_end = clock();
    if (clock_end < clock_begin) {
        my_printf("clock time cycle detected!\n");
        clock_begin = clock_end;
    } else {
        if ((clock_end - clock_begin) > (CLOCKS_PER_SEC / 10)) {
            time_used += (double) (clock_end - clock_begin) / CLOCKS_PER_SEC;
            clock_begin = clock_end;
            if (time_used > stop_time)
                return (SCORE_INVAL);
        }
    }

    temp_score = alpha;
    alpha = -beta;
    beta = -temp_score;

    base = capt_top;

    best_score = -SCORE_LIMIT;
    best_move = INVALID;

    p = cache[my_colour] + (node->node_total_key.low & CACHE_MASK);
    if (!(p->cache_key.low & CACHE_INVAL) and
        (p->cache_key.high == node->node_total_key.high) and
#ifdef HASH_KEY128
        (p->cache_key.lower == node->node_total_key.lower) and
        (p->cache_key.higher == node->node_total_key.higher) and
#endif
        ((p->cache_key.low & ~CACHE_KIND_MASK) == (node->node_total_key.low & ~CACHE_KIND_MASK))) {
        best_move = p->cache_best;
        if ((p->cache_key.low & CACHE_DEPTH_MASK) >= depth) {
            if (p->cache_key.low & CACHE_true)
			{
                RETURN(p->cache_score);
			}
            else if (p->cache_key.low & CACHE_LOWER) {
                if (p->cache_score >= beta)
				{
                    RETURN(p->cache_score);
				}
                else if (p->cache_score > alpha)
                    alpha = p->cache_score;
            } else if (p->cache_key.low & CACHE_UPPER) {
                if (p->cache_score <= alpha)
				{
                    RETURN(p->cache_score);
				}
                else if (p->cache_score < beta)
                    beta = p->cache_score;
            }
        }
    }
    if (draw_by_repetition(false))
        RETURN(0);
    check = i_am_check(my_colour, your_colour);

    if ((depth == 0) and ! check) {
        best_score = 0;
        goto label_store;
    }
    nlegal = 0;
    ngens++;
    nmoves = gen_all_moves();

    if (depth == 0) {
        for (imove = 0; imove < nmoves; imove++) {
            do_move(base[imove]);
            if (i_am_check(your_colour, my_colour)) {
                undo_move();
                continue;
            } else {
                undo_move();
                best_score = 0;
                goto label_store;
            }
        }
        best_score = -SCORE_MATE + inode;
        goto label_store;
    }
    if (best_move == INVALID)
        best_move = 0;
    else if (best_move >= nmoves) {
        my_printf("hash fault!\n");
        best_move = 0;
    }
    for (imove = INVALID; imove < nmoves; imove++) {
        if (imove == INVALID)
            jmove = best_move;
        else
            jmove = imove;
        if (base[jmove] == INVAL_MOVE)
            continue;
        do_move(base[jmove]);
        if (i_am_check(your_colour, my_colour)) {
            undo_move();
            base[jmove] = INVAL_MOVE;
            continue;
        }
        ndos++;
        nlegal++;


        /* Another short cut. The move at depth 1 has to be a checking move. */

        if ((depth == 1) and ! i_am_check(my_colour, your_colour))        // - JA
         temp_score = 0;
        else
            temp_score = mate_alpha_beta(MAX(alpha, best_score), beta, depth - 1, var_top);
        undo_move();
        if (temp_score == SCORE_INVAL)
            return (SCORE_INVAL);
        if (temp_score > best_score) {
            best_score = temp_score;
            best_move = jmove;
            var = var_base;
            *var++ = base[jmove];
            while ((*var++ = *var_top++) != INVAL_MOVE);
            var_top = var;
            var = var_base;
            if (best_score >= beta)
                goto label_store;
        }
        base[jmove] = INVAL_MOVE;
    }

    if (nlegal == 0) {
        if (check)
		{
            RETURN(-SCORE_MATE)+ inode; //JA
		}
        else
		{
            RETURN(0);
		}
    }
label_store:

    if (depth > CACHE_DEPTH_MASK)
        goto label_neg;
    p->cache_key.high = node->node_total_key.high;

    p->cache_key.low = ((node->node_total_key.low & ~CACHE_KIND_MASK) | depth);
    p->cache_score = best_score;
    if (best_score <= alpha)
        p->cache_key.low |= CACHE_UPPER;
    else if (best_score >= beta)
        p->cache_key.low |= CACHE_LOWER;
    else
        p->cache_key.low |= CACHE_true;
    p->cache_best = best_move;

label_neg:
    BUG(best_score == -SCORE_LIMIT);
    capt_top = base;
    return (-best_score);
#undef RETURN
}
/*
   The legal moves in the starting position are stored in array root.
 */

int             nroot;
root_t         *root;
int             jroot;

/*
   gen_root finds the legal moves in the starting position.
 */

void            gen_root(void)
{
    int             imove,
                    nmoves;
    int             check;
    move_t         *base;

    nroot = 0;
    capt_top = capt_stack;
    base = capt_top;
    nmoves = gen_all_moves();
    for (imove = 0; imove < nmoves; imove++) {
        do_move(base[imove]);
        check = i_am_check(your_colour, my_colour);
        undo_move();
        if (check)
            continue;
        BUG(nroot >= NROOT_MAX);
        root[nroot].root_move = base[imove];
        nroot++;
    }
    for (imove = 0; imove < nroot; imove++)
        root[imove].root_score = 0;
    jroot = 0;
}

/*
   search_root checks whether move occurs in the list of root moves.
   Returns INVALID if the move is not found.
   Prints the legal moves if the move is not found and the verbose flag is set.
 */

static int             search_root(char *move, int option, int strict, int verbose)
{
    int             iroot;
    int             len;

    len = strlen(move);
    for (iroot = 0; iroot < nroot; iroot++) {
        if (option != MOVE_STRING_LONG) {
            move_to_san_string(root[iroot].root_move, true);
            if (strict) {
                if (strcmp(move, move_string) == 0)
                    break;
            } else {
                if (strncmp(move_string, move, len) == 0)
                    break;
            }
        }
        if (option != MOVE_STRING_SAN) {
            move_to_string(root[iroot].root_move);
            if (strict) {
                if (strcmp(move, move_string) == 0)
                    break;
            } else {
                if (strncmp(move_string, move, len) == 0)
                    break;
            }
        }
    }
    if (iroot < nroot)
        return (iroot);
    if (verbose) {
        
        
                            my_printf("move %s\n", move);
        
        for (iroot = 0; iroot < nroot; iroot++) {
            move_to_san_string(root[iroot].root_move, true);
            my_printf(" %7s", move_string);
            if (((iroot % 9) == 8) or(iroot == (nroot - 1)))
                my_printf("\n");
        }
    }
    return (INVALID);
}

/*
   display_var displays the best variation.
 */



static void      clear_cache(void)
{
    int             icolour;
    long            i;

    for (icolour = WHITE_COLOUR; icolour <= BLACK_COLOUR; icolour++) {
        for (i = 0; i < (CACHE_MASK + 1); i++)
            cache[icolour][i].cache_key.low = CACHE_INVAL;
        for (i = 0; i < (SCORE_CACHE_MASK + 1); i++)
            score_cache[icolour][i].cache_score = SCORE_INVAL;
    }
}


static void      display_var(move_t best_move, var_t * var_base)
{
    move_to_san_string(best_move, true);
    my_printf(" %s", move_string);
    if (draw_by_repetition(false)) return;
    do_move(best_move);
    if (i_am_check(your_colour, my_colour)) {
        my_printf("hash fault, check!\n");
        
        move_to_string(best_move);
        my_printf("%s%s\n", colour_as_string[my_colour], move_string);
    } else {
        if (*var_base != INVAL_MOVE)
            display_var(*var_base, var_base + 1);
    }
    undo_move();
}



static int             do_mate(void)
{
    int             iroot,
                    depth;
    int             temp_score;
    int newdepth;

    if (ZapCache)
        clear_cache();

    clock_begin = clock();
    time_used = 0;
    inode_root = inode;
    inode_max = 0;

    for (depth = 1; depth <= 2 * MateDepth - 1; depth += 2) {
        ngens = ndos = 0;

        for (iroot = 0; iroot < nroot; iroot++) {
            do_move(root[iroot].root_move);
            temp_score = mate_alpha_beta(SCORE_ABORT - 1, SCORE_ABORT, depth - 1, var_stack);
            undo_move();
            if (temp_score == SCORE_INVAL)
                return (FALSE);
            if (temp_score > SCORE_ABORT) {
                display_var(root[iroot].root_move, var_stack);
                my_printf(" %.3f\n", (double) temp_score);
                return (TRUE);
            }
        }
        
     //    if (depth > newdepth) 
         my_printf("%d %.f %.0f %d ",depth,(double) root[jroot].root_score/5, time_used, ndos/100*5);
            display_var(root[jroot].root_move, var_stack);
            printf("\n");
       
       //  newdepth = depth;
       
        my_printf("depth=%d time=%.2f stats=%ld/%ld\n",
                  depth, time_used, ngens, ndos);
    }
    return (FALSE);
}



/*
   do_root tries to find the best move by iterative deepening,
   aspiration search and minimal window search.
 */

static void            do_root(void)
{
    int             check;
    int             depth;
    int             newdepth;
    score_t         temp_score,
                    fh_score;
    int             depth_next;
    score_t         alpha,
                    beta;
    int             iroot;
   

   
    
    if (ZapCache)
        clear_cache();
    check = i_am_check(my_colour, your_colour);

    jroot = 0;
    root[0].root_score = node->node_mat[my_colour] - node->node_mat[your_colour];
    inode_root = inode;
    inode_max = 0;
    
    clock_begin = clock();
    time_used = 0;

    iroot = 0;
    temp_score = SCORE_INVAL;

    /* Iterative deepening. */


    for (depth = 1; depth <= DepthMax; depth++) {
#ifdef BUG_DEPTH
        {
            int             ibug;
            for (ibug = 0; ibug < NODE_MAX; ibug++)
                bug_depth[ibug] = 0;
        }
#endif

        depth_next = (check ? depth : depth - 1);
        ngens = ndos = 0;
        nevals = nhits = 0;


        /* Aspiration search with half a pawn margin. */
 
        alpha = root[jroot].root_score - MARGIN;
        beta = root[jroot].root_score + MARGIN;      

        do_move(root[jroot].root_move);
        temp_score = alpha_beta(false, alpha, beta, depth_next, var_stack);
        undo_move();
        if (temp_score == SCORE_INVAL)
            goto label_final;

        /* Research if fail low or fail high. */


        do_move(root[jroot].root_move);
        if (temp_score <= alpha)
            temp_score = alpha_beta(false, -SCORE_LIMIT, temp_score, depth_next, var_stack);
        else if (temp_score >= beta)
            temp_score = alpha_beta(false, temp_score, SCORE_LIMIT, depth_next, var_stack);
        undo_move();
        if (temp_score == SCORE_INVAL)
            goto label_final;
        root[jroot].root_score = temp_score;
               
      //  if (depth > newdepth) 
        my_printf("%d %.f %.0f %d ",depth,(double) root[jroot].root_score/5, time_used, ndos/100*5);
            display_var(root[jroot].root_move, var_stack);
            printf("\n");
        /* Search the remaining moves with a minimal window. */
      //   newdepth = depth;

        for (iroot = 0; iroot < nroot; iroot++) {
            if (iroot == jroot)
                continue;

            alpha = root[jroot].root_score;
            beta = alpha + 1;

               
            do_move(root[iroot].root_move);
            temp_score = alpha_beta(false, alpha, beta, depth_next, var_stack);
            undo_move();
            
        
        
            
            if (temp_score == SCORE_INVAL) {
               
                goto label_final;
                }
            /* Research if fail high. */

            if (temp_score > beta) {
               
                jroot = iroot;
                root[jroot].root_score = temp_score;
                        
                printf("%d", depth);
               
                display_var(root[jroot].root_move, var_stack);
                my_printf(" %.f\n", (double) temp_score);
                do_move(root[jroot].root_move);

                fh_score = temp_score;
                temp_score = alpha_beta(false, temp_score, SCORE_LIMIT, depth_next, var_stack);
                undo_move();
                if (temp_score == SCORE_INVAL)
                    goto label_final;

                /* If the reseach fails low, there was a null move failure
                 * (alas), a (depth >= cache_depth) failure (see alpha beta)
                 * or a hash problem. */
                if (temp_score < fh_score)
                //    my_printf("research error!\n");
                root[jroot].root_score = temp_score;
                
            
          //  if (depth > newdepth) 
            my_printf("%d %.f %.0f %d ", depth, (double) root[jroot].root_score/5, time_used, ndos/100*5);
            display_var(root[jroot].root_move, var_stack);
            printf("\n");
            
          //   newdepth = depth;
            
            }
        }
#ifdef BUG_DEPTH
        {
            int             ibug;
            for (ibug = 0; ibug < NODE_MAX; ibug++)
                if (bug_depth[ibug] != 0)
                    my_printf("%d %d\n", ibug, bug_depth[ibug]);
        }
#endif

                   
                 
                        
                        
                        
        if (ABS(root[jroot].root_score) > SCORE_ABORT)
            goto label_final;
    }
label_final:

    /* Only print this if time limit. */

    if (temp_score == SCORE_INVAL) {
        
        
      //   if (depth > newdepth) 
         my_printf("%d %.f %.0f %d ",
                  depth,(double) root[jroot].root_score/5, time_used, ndos/100*5);
                 printf("\n"); 
                    
             //  newdepth = depth;    
    }
    if (jroot == INVALID) jroot = 0;
    
   
Best_Score = root[jroot].root_score;
}

/*
   Initialization of some arrays.
 */

static void            initialize(void)
{
    int             i,
                    j,
                    n;
    int             icolour,
                    ikind;


    /* The bit count arrays. */

    MALLOC(bit_count, int, NBITS_SIZE);
    MALLOC(bit_low, int, NBITS_SIZE);
    MALLOC(bit_high, int, NBITS_SIZE);
    for (i = 0; i < NBITS_SIZE; i++) {
        n = 0;
        for (j = 0; j < NBITS; j++)
            if ((i & BIT(j)) != 0)
                n++;
        bit_count[i] = n;
        if (i == 0)
            bit_low[i] = bit_high[i] = INVALID;
        else {
            for (j = 0; j < NBITS; j++)
                if ((i & BIT(j)) != 0)
                    break;
            bit_low[i] = j;
            for (j = NBITS - 1; j >= 0; j--)
                if ((i & BIT(j)) != 0)
                    break;
            bit_high[i] = j;
        }
    }
    MALLOC(row, int, BOARD_MAX);
    MALLOC(column, int, BOARD_MAX);
    for (i = 0; i < BOARD_MAX; i++)
        row[i] = column[i] = -1;
    for (i = 0; i < 64; i++) {
        row[map[i]] = row_table[i];
        column[map[i]] = column_table[i];
    }
    MALLOC(nota, char *, BOARD_MAX);
    for (i = 0; i < BOARD_MAX; i++)
        nota[i] = nota_inval;
    for (i = 0; i < 64; i++)
        nota[map[i]] = nota_table[i];
    MALLOC(move_stack, move_t, MOVE_MAX);
    MALLOC(capt_stack, move_t, CAPT_MAX);
    MALLOC(var_stack, var_t, VAR_MAX);
    move_top = move_stack;
    capt_top = capt_stack;
    MALLOC(node_stack, node_t, NODE_MAX);
    node_stack[0].node_reverse = false;
    node_stack++;
    node = node_stack;
    inode = inode_max = 0;

    my_printf("size of hash tables: %lu %lu\n",
              (ulong_t) sizeof(cache_t) * CACHE_MASK * 2,
              (ulong_t) sizeof(score_cache_t) * SCORE_CACHE_MASK * 2);

    HUGE_MALLOC(cache[WHITE_COLOUR], cache_t, CACHE_MASK + 1);
    HUGE_MALLOC(cache[BLACK_COLOUR], cache_t, CACHE_MASK + 1);
    HUGE_MALLOC(score_cache[WHITE_COLOUR], score_cache_t, SCORE_CACHE_MASK + 1);
    HUGE_MALLOC(score_cache[BLACK_COLOUR], score_cache_t, SCORE_CACHE_MASK + 1);

    clear_cache();

    /* The hash keys. */

    for (icolour = WHITE_COLOUR; icolour <= BLACK_COLOUR; icolour++) {
        for (ikind = PAWN; ikind <= KING; ikind++) {

            HUGE_MALLOC(hash_key[icolour][ikind], hash_key_t, BOARD_MAX);

            for (i = 0; i < 64; i++)
                SET_KEY(hash_key[icolour][ikind][map[i]]);
        }
    }

    HUGE_MALLOC(ep_key, hash_key_t, BOARD_MAX);
    HUGE_MALLOC(castle_key, hash_key_t, BOARD_MAX);

    for (i = 0; i < 64; i++) {
        SET_KEY(ep_key[map[i]]);
        SET_KEY(castle_key[map[i]]);
    }

    /* Some more tables. */

    MALLOC(king_alone_grad, int, BOARD_MAX);
    MALLOC(white_free_pawn_grad, int, BOARD_MAX);
    MALLOC(black_free_pawn_grad, int, BOARD_MAX);
    free_pawn_grad[WHITE_COLOUR] = white_free_pawn_grad;
    free_pawn_grad[BLACK_COLOUR] = black_free_pawn_grad;

    MALLOC(white_pawn_visit, int, BOARD_MAX);
    MALLOC(black_pawn_visit, int, BOARD_MAX);
    pawn_visit[WHITE_COLOUR] = white_pawn_visit;
    pawn_visit[BLACK_COLOUR] = black_pawn_visit;

    MALLOC(white_knight_visit, int, BOARD_MAX);
    MALLOC(black_knight_visit, int, BOARD_MAX);
    knight_visit[WHITE_COLOUR] = white_knight_visit;
    knight_visit[BLACK_COLOUR] = black_knight_visit;

    MALLOC(pawn_board, int, BOARD_MAX);
    for (i = 0; i < BOARD_MAX; i++) {
        king_alone_grad[i] = 0;
        white_free_pawn_grad[i] = black_free_pawn_grad[i] = 0;
        white_pawn_visit[i] = black_pawn_visit[i] = 0;
        white_knight_visit[i] = black_knight_visit[i] = 0;
    }
    for (i = 0; i < 64; i++) {
        king_alone_grad[map[i]] = king_alone_grad_table[i];
        white_free_pawn_grad[map[i]] = free_pawn_grad_table[i];
        black_free_pawn_grad[mirror_map[i]] = free_pawn_grad_table[i];
        white_pawn_visit[map[i]] = pawn_visit_table[i];
        black_pawn_visit[mirror_map[i]] = pawn_visit_table[i];
        white_knight_visit[map[i]] = knight_visit_table[i];
        black_knight_visit[mirror_map[i]] = knight_visit_table[i];
    }

    /* Scale the free pawn gradient such that a pawn on rank 7 is worth a
     * pawn */

    for (i = 0; i < 64; i++) {
        white_free_pawn_grad[map[i]] *= 3;
        black_free_pawn_grad[map[i]] *= 3;
    }
    MALLOC(attack_board, int, BOARD_MAX);
    MALLOC(root, root_t, NROOT_MAX);
  
}







       
